# -*- coding: utf-8 -*-
"""
nidaba.tasks
============

The tasks package contains all built-in celery tasks used by nidaba. Tasks
should not contain any actual implementations but import them as separate
packages and just shim around them doing some initial setup, most commonly
conversion between storage module tuples and absolute paths.

Additional tasks depending on external or complex dependencies are contained in
plugins found in the plugins directory.
"""

from __future__ import absolute_import

from nidaba.tasks import util
from nidaba.tasks import img
from nidaba.tasks import binarize
from nidaba.tasks import stats
from nidaba.tasks import postprocessing
from nidaba.tasks import output



"""
nidaba.image
~~~~~~~~~~~~

Common image processing functions encapsulating the PIL or pythonica image
interface to absolute file paths.
"""

from __future__ import absolute_import

from PIL import Image

import nidaba.algorithms.otsu


def otsu(imagepath, resultpath):
    """
    Binarizes an grayscale image using Otsu's algorithm.

    Arguments:
        imagepath: Path of the input image
        resultpath: Path of the output image

    Returns:
        unicode: Path of the actual output file
    """

    im = Image.open(imagepath)
    nidaba.algorithms.otsu.otsu(im).save(resultpath)
    return resultpath


def rgb_to_gray(imagepath, resultpath):
    """
    Converts an RGB or CMYK image into a 8bpp grayscale image.

    Arguments:
        imagepath: Path of the input image
        resultpath: Path of the output image

    Returns:
        unicode: Path of the actual output file
    """

    img = Image.open(imagepath)
    img.convert('L').save(resultpath)
    return resultpath



# -*- coding: utf-8 -*-
"""
nidaba.plugins.kraken
~~~~~~~~~~~~~~~~~~~~~

Plugin implementing access to kraken functions.

`kraken <https://mittagessen.github.io/kraken>`_ is a fork of OCRopus
implementing sane interfaces while preserving (mostly) functional equivalence.
To use this plugin kraken has to be installed into the current python path,
e.g. the current virtualenv. It is available from pypi::

    $ pip install kraken

It should be able to utilize any model trained for ocropus and is configured
using the same global configuration options.
"""

from __future__ import absolute_import

import os
import shutil

from nidaba import uzn
from nidaba import storage
from nidaba.tei import TEIFacsimile
from nidaba.config import nidaba_cfg
from nidaba.celery import app
from nidaba.nidabaexceptions import NidabaInvalidParameterException
from nidaba.nidabaexceptions import NidabaPluginException
from nidaba.tasks.helper import NidabaTask

from PIL import Image


def setup(*args, **kwargs):
    try:
        global binarization
        global pageseg
        global rpred
        global models
        from kraken import binarization
        from kraken import pageseg
        from kraken import rpred
        from kraken.lib import models
    except ImportError as e:
        raise NidabaPluginException(e.message)


@app.task(base=NidabaTask, name=u'nidaba.segmentation.kraken')
def segmentation_kraken(doc, method=u'segment_kraken', black_colseps=False):
    """
    Performs page segmentation using kraken's built-in algorithm and writes a
    skeleton TEI file.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string append to all output files
        black_colseps (bool): Assume black column separator instead of white
        ones.

    Returns:
        Two storage tuples with the first one containing the segmentation and
        the second one being the file the segmentation was calculated upon.
    """

    input_path = storage.get_abs_path(*doc)
    output_path, ext = os.path.splitext(storage.insert_suffix(input_path,
                                        method))
    shutil.copy2(input_path, output_path + ext)
    img = Image.open(input_path)
    with open(output_path + '.xml', 'w') as fp:
        tei = TEIFacsimile()
        tei.document(img.size, os.path.join(*doc))
        tei.title = os.path.basename(doc[1])
        tei.add_respstmt('kraken', 'page segmentation')
        for seg in pageseg.segment(img, black_colseps):
            tei.add_line(seg)
        tei.write(fp)
    return (storage.get_storage_path(output_path + '.xml'),
            storage.get_storage_path(output_path + ext))


@app.task(base=NidabaTask, name=u'nidaba.ocr.kraken')
def ocr_kraken(doc, method=u'ocr_kraken', model=None):
    """
    Runs kraken on an input document and writes a TEI file.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string append to all output files
        model (unicode): Identifier for the font model to use

    Returns:
        (unicode, unicode): Storage tuple for the output file
    """
    input_path = storage.get_abs_path(*doc[1])
    output_path = (doc[1][0], os.path.splitext(storage.insert_suffix(doc[1][1],
                                                                     method,
                                                                     model))[0]
                   + '.xml')
    segmentation = storage.get_abs_path(*doc[0])
    if model in nidaba_cfg['kraken_models']:
        model = storage.get_abs_path(*(nidaba_cfg['kraken_models'][model]))
    elif model in nidaba_cfg['ocropus_models']:
        model = storage.get_abs_path(*(nidaba_cfg['ocropus_models'][model]))
    else:
        raise NidabaInvalidParameterException('Model not defined in '
                                              'configuration')
    img = Image.open(input_path)
    tei = TEIFacsimile()
    with open(segmentation, 'r') as seg:
        tei.read(seg)
    # kraken is a line recognizer
    tei.clear_graphemes()
    tei.clear_segments()
    # add and scope new responsibility statement
    tei.add_respstmt('kraken', 'character recognition')
    lines = tei.lines

    rnn = models.load_any(model)
    i = 0
    for rec in rpred.rpred(rnn, img, [(int(x[0]), int(x[1]), int(x[2]), int(x[3])) for x in lines]):
        # scope the current line and add all graphemes recognized by kraken to
        # it.
        tei.scope_line(lines[i][4])
        tei.add_graphemes(rec)
        i += 1
    with open(storage.get_abs_path(*output_path), 'w') as fp:
        tei.write(fp)
    return output_path


@app.task(base=NidabaTask, name=u'nidaba.binarize.nlbin')
def nlbin(doc, method=u'nlbin', threshold=0.5, zoom=0.5, escale=1.0,
          border=0.1, perc=80, range=20, low=5, high=90):
    """
    Binarizes an input document utilizing ocropus'/kraken's nlbin algorithm.

    Args:
        doc (unicode, unicode): The input document tuple.
        method (unicode): The suffix string appended to all output files.
        threshold (float):
        zoom (float):
        escale (float):
        border (float)
        perc (int):
        range (int):
        low (int):
        high (int):

    Returns:
        (unicode, unicode): Storage tuple of the output file

    Raises:
        NidabaInvalidParameterException: Input parameters are outside the valid
                                         range.

    """
    input_path = storage.get_abs_path(*doc)
    output_path = storage.insert_suffix(input_path, method, unicode(threshold),
                                        unicode(zoom), unicode(escale),
                                        unicode(border), unicode(perc),
                                        unicode(range), unicode(low),
                                        unicode(high))
    kraken_nlbin(input_path, output_path, threshold, zoom, escale, border,
                 perc, range, low, high)
    return storage.get_storage_path(output_path)


def kraken_nlbin(input_path, output_path, threshold=0.5, zoom=0.5, escale=1.0,
                 border=0.1, perc=80, range=20, low=5, high=90):
    """
    Binarizes an input document utilizing ocropus'/kraken's nlbin algorithm.

    Args:
        input_path (unicode): Path to the input image
        output_path (unicode): Path to the output image
        threshold (float):
        zoom (float):
        escale (float):
        border (float)
        perc (int):
        range (int):
        low (int):
        high (int):

    Raises:
        NidabaInvalidParameterException: Input parameters are outside the valid
                                         range.

    """
    img = Image.open(input_path)
    binarization.nlbin(img, threshold, zoom, escale, border, perc, range, low,
                       high).save(output_path)



# -*- coding: utf-8 -*-
#
# nidaba documentation build configuration file, created by
# sphinx-quickstart on Tue Nov  4 18:48:48 2014.

from subprocess import Popen, PIPE

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'celery.contrib.sphinx',
]

templates_path = ['_templates']

source_suffix = '.rst'

master_doc = 'index'

project = u'nidaba'
copyright = u'2014-2015, Open Greek and Latin'

pipe = Popen('git describe --tags --always master', stdout=PIPE, shell=True)
version = pipe.stdout.read()
release = version

exclude_patterns = ['_build']

html_theme = 'alabaster'
html_static_path = ['_static']
htmlhelp_basename = 'nidabadoc'

html_theme_options = {
    'description': "An expandable and scalable OCR pipeline",
    'github_user': 'openphilology',
    'github_repo': 'nidaba',
    'travis_button': 'OpenPhilology/nidaba',
}


html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'searchbox.html',
        'donate.html',
    ]
}

pygments_style = 'sphinx'
latex_elements = {}
latex_documents = [('index', 'nidaba.tex', u'nidaba Documentation', u'Open\
                     Greek and Latin', 'manual')]
man_pages = [('index', 'nidaba', u'nidaba Documentation', [u'Open Greek and\
                                                            Latin'], 1)]
texinfo_documents = [('index', 'nidaba', u'nidaba Documentation', u'Open Greek\
                       and Latin', 'nidaba', 'Expandable and scalable OCR\
                       pipeline', 'Miscellaneous')]



# -*- coding: utf-8 -*-
import unittest
import os
import shutil
import tempfile

from nose.plugins.skip import SkipTest
from mock import patch, MagicMock

thisfile = os.path.abspath(os.path.dirname(__file__))


class KrakenTests(unittest.TestCase):

    """
    Tests for the kraken plugin.
    """

    def setUp(self):
        config_mock = MagicMock()
        storage_path = unicode(tempfile.mkdtemp())
        config_mock.nidaba_cfg = {
            'storage_path': storage_path,
        }

        self.patches = {
            'nidaba.config': config_mock,
        }
        self.patcher = patch.dict('sys.modules', self.patches)
        self.addCleanup(self.patcher.stop)
        self.patcher.start()
        self.storage_path = storage_path
        from nidaba import storage
        self.storage = storage

    def tearDown(self):
        self.patcher.stop()
        shutil.rmtree(self.storage_path)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
nidaba.plugins.tesseract
~~~~~~~~~~~~~~~~~~~~~~~~

Plugin implementing an interface to tesseract

This plugin exposes tesseract's functionality as a task. It implements two ways
of calling tesseract, a direct method calling the tesseract executable and one
utilizing the C-API available from tesseract 3.02 and upwards.

The C-API requires a libtesseract shared object in the current library path and
training data in the configured tessdata directory:

.. code-block:: console

    # apt-get install libtesseract3 tesseract-ocr-$lang

Using the direct call method requires the tesseract binary installable by
executing:

.. code-block:: console

    # apt-get install tesseract-ocr

.. note::
    It is strongly encouraged to use the C-API whenever possible. It is
    supposedly stable while hOCR output file names change between tesseract
    versions.

.. note::
    Parameters in configuration files supersede command line parameters.
    Modular page segmentation utilizing zone files requires that the page
    segmentation mode may be set freely. Uncomment the line:

        tessedit_pageseg_mode 1

    in the default hocr configuration (in TESSDATA/configs/).

Configuration
~~~~~~~~~~~~~

implementation (default='capi')
    Selector for the call method. May either be `capi`, `direct` (tesseract
    hOCR output with .hocr extension), or `legacy` (tesseract hOCR output with
    .html extension).

tessdata (default='/usr/share/tesseract-ocr/')
    Path to load tesseract training data and configuration from. Has to be one
    directory level upwards from the actual tessdata directory
"""

from __future__ import absolute_import

import subprocess
import StringIO
import ctypes
import os

from PIL import Image
from shutil import copyfile
from distutils import spawn
from os.path import splitext
from celery.utils.log import get_task_logger

from nidaba.uzn import UZNWriter
from nidaba import storage
from nidaba.tei import TEIFacsimile
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask
from nidaba.nidabaexceptions import NidabaTesseractException
from nidaba.nidabaexceptions import NidabaPluginException


implementation = u'capi'
tessdata = u'/usr/share/tesseract-ocr/'

(RIL_BLOCK, RIL_PARA, RIL_TEXTLINE, RIL_WORD, RIL_SYMBOL) = map(ctypes.c_int,
                                                                xrange(5))

logger = get_task_logger(__name__)


class TessBaseAPI(ctypes.Structure):
    """
    Dummy class encapsulating the TessBaseAPI struct returned by
    TessBaseAPICreate().
    """
    pass


def setup(*args, **kwargs):
    if kwargs.get(u'implementation'):
        global implementation
        implementation = kwargs.get(u'implementation')
    if kwargs.get(u'tessdata'):
        global tessdata
        if isinstance(kwargs.get(u'tessdata'), list):
            tessdata = storage.get_abs_path(*kwargs.get(u'tessdata'))
        else:
            tessdata = kwargs.get(u'tessdata')
    if implementation == 'direct' and not spawn.find_executable('tesseract'):
        raise NidabaPluginException('No tesseract executable found')
    if implementation == 'capi':
        try:
            ctypes.cdll.LoadLibrary('libtesseract.so.3')
        except:
            raise NidabaPluginException('Loading libtesseract failed.')


@app.task(base=NidabaTask, name=u'nidaba.segmentation.tesseract')
def segmentation_tesseract(doc, method=u'segment_tesseract'):
    """
    Performs page segmentation using tesseract's built-in algorithm and writes
    a TEI XML segmentation file.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files.

    Returns:
        Two storage tuples with the first one containing the segmentation and
        the second one being the file the segmentation was calculated upon.
    """
    input_path = storage.get_abs_path(*doc)
    output_path = splitext(storage.insert_suffix(input_path, method))[0] + '.xml'

    try:
        tesseract = ctypes.cdll.LoadLibrary('libtesseract.so.3')
    except OSError as e:
        raise NidabaTesseractException('Loading libtesseract failed: ' +
                                       e.message)
    tesseract.TessVersion.restype = ctypes.c_char_p
    tesseract.TessBaseAPICreate.restype = ctypes.POINTER(TessBaseAPI)
    ver = tesseract.TessVersion()
    if int(ver.split('.')[0]) < 3 or int(ver.split('.')[1]) < 2:
        raise NidabaTesseractException('libtesseract version is too old. Set '
                                       'implementation to direct.')
    api = tesseract.TessBaseAPICreate()
    rc = tesseract.TessBaseAPIInit3(api, tessdata.encode('utf-8'), None)
    if (rc):
        tesseract.TessBaseAPIDelete(api)
        raise NidabaTesseractException('Tesseract initialization failed.')

    # only do segmentation and script detection
    tesseract.TessBaseAPISetPageSegMode(api, 4)
    tesseract.TessBaseAPIProcessPages(api, input_path.encode('utf-8'), None, 0, None)
    it = tesseract.TessBaseAPIAnalyseLayout(api)
    x0, y0, x1, y1 = (ctypes.c_int(), ctypes.c_int(), ctypes.c_int(),
                      ctypes.c_int())

    # initialize XML file
    tei = TEIFacsimile()
    tei.document(Image.open(input_path).size, os.path.join(*doc))
    tei.title = os.path.basename(doc[1])
    tei.add_respstmt('tesseract', 'page segmentation')

    while True:
        if tesseract.TessPageIteratorIsAtBeginningOf(it, RIL_TEXTLINE):
            tesseract.TessPageIteratorBoundingBox(it,
                                                  RIL_TEXTLINE,
                                                  ctypes.byref(x0),
                                                  ctypes.byref(y0),
                                                  ctypes.byref(x1),
                                                  ctypes.byref(y1))
            tei.add_line((x0.value, y0.value, x1.value, y1.value))
        if tesseract.TessPageIteratorIsAtBeginningOf(it, RIL_WORD):
            tesseract.TessPageIteratorBoundingBox(it,
                                                  RIL_WORD,
                                                  ctypes.byref(x0),
                                                  ctypes.byref(y0),
                                                  ctypes.byref(x1),
                                                  ctypes.byref(y1))
            tei.add_segment((x0.value, y0.value, x1.value, y1.value))
        tesseract.TessPageIteratorBoundingBox(it,
                                              RIL_SYMBOL,
                                              ctypes.byref(x0),
                                              ctypes.byref(y0),
                                              ctypes.byref(x1),
                                              ctypes.byref(y1))
        tei.add_graphemes([(None, (x0.value, y0.value, x1.value, y1.value))])
        if not tesseract.TessPageIteratorNext(it, RIL_SYMBOL):
            break
    tesseract.TessPageIteratorDelete(it)
    tesseract.TessBaseAPIDelete(api)
    with open(output_path, 'w') as fp:
        tei.write(fp)
    return storage.get_storage_path(output_path), doc


@app.task(base=NidabaTask, name=u'nidaba.ocr.tesseract')
def ocr_tesseract(doc, method=u'ocr_tesseract', languages=None,
                  extended=False):
    """
    Runs tesseract on an input document.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files
        languages (list): A list of tesseract classifier identifiers
        extended (bool): Switch to enable extended hOCR generation containing
                         character cuts and confidences. Has no effect when
                         direct or legacy implementation is used.

    Returns:
        (unicode, unicode): Storage tuple for the output file
    """
    image_path = storage.get_abs_path(*doc[1])

    # rewrite the segmentation file to lines in UZN format
    seg = TEIFacsimile()
    with open(storage.get_abs_path(*doc[0])) as fp:
        seg.read(fp)
    with open(splitext(image_path)[0] + '.uzn', 'w') as fp:
        uzn = UZNWriter(fp)
        for line in seg.lines:
            uzn.writerow(*tuple(int(l) for l in line[:-2]))

    if isinstance(languages, basestring):
        languages = [languages]
    output_path = storage.insert_suffix(image_path, method, *languages)

    if implementation == 'legacy':
        result_path = output_path + '.html'
        ocr_direct(image_path, output_path, languages)
    elif implementation == 'direct':
        result_path = output_path + '.hocr'
        ocr_direct(image_path, output_path, languages)
    elif implementation == 'capi':
        result_path = output_path + '.xml'
        ocr_capi(image_path, result_path, seg, languages, extended)
    else:
        raise NidabaTesseractException('Invalid implementation selected',
                                       implementation)
    if not result_path[-4:] == '.xml':
        tei = TEIFacsimile()
        with open(result_path) as fp:
            tei.load_hocr(fp)
        os.unlink(result_path)
        with open(output_path + '.xml', 'wb') as fp:
            tei.write(fp)
        result_path = output_path + '.xml'
    return storage.get_storage_path(result_path)


def ocr_capi(image_path, output_path, facsimile, languages, extended=False):
    """
    OCRs an image using the C API provided by tesseract versions 3.02 and
    higher.

    Args:
        image_path (unicode): Path to the input image
        facsimile (nidaba.tei.TEIFacsimile): Facsimile object of the
                                             segmentation
        output_path (unicode): Path to the hOCR output
        languages (list): List of valid tesseract language identifiers
        extended (bool): Switch to select extended hOCR output containing
                         character cuts and confidences values
    """

    try:
        tesseract = ctypes.cdll.LoadLibrary('libtesseract.so.3')
    except OSError as e:
        raise NidabaTesseractException('Loading libtesseract failed: ' +
                                       e.message)

    # set up all return types
    tesseract.TessVersion.restype = ctypes.c_char_p
    tesseract.TessBaseAPICreate.restype = ctypes.POINTER(TessBaseAPI)
    if extended:
        try:
            tesseract.TessResultIteratorConfidence.restype = ctypes.c_float
            tesseract.TessResultIteratorWordRecognitionLanguage.restype = ctypes.c_char_p
            tesseract.TessResultIteratorGetUTF8Text.restype = ctypes.c_char_p
        except AttributeError as e:
            raise NidabaTesseractException('Symbols required for extended '
                    'output not available. Rerun using standard output.')
    # ensure we've loaded a tesseract object newer than 3.02
    ver = tesseract.TessVersion()
    if int(ver.split('.')[0]) < 3 or int(ver.split('.')[1]) < 2:
        raise NidabaTesseractException('libtesseract version is too old. Set '
                                       'implementation to direct.')
    api = tesseract.TessBaseAPICreate()
    rc = tesseract.TessBaseAPIInit3(api, tessdata.encode('utf-8'),
                                    ('+'.join(languages)).encode('utf-8'))
    if (rc):
        tesseract.TessBaseAPIDelete(api)
        raise NidabaTesseractException('Tesseract initialization failed.')

    tesseract.TessBaseAPISetPageSegMode(api, 3)

    tesseract.TessBaseAPIProcessPages(api, image_path.encode('utf-8'), None, 0, None)
    if tesseract.TessBaseAPIRecognize(api, None):
        tesseract.TessBaseAPIDelete(api)
        raise NidabaTesseractException('Tesseract recognition failed')
    if extended:
        facsimile.add_respstmt('tesseract', 'character recognition')
        # While tesseract can recognize single words/characters it is extremely
        # slow to do so. We therefore wrote an UZN file containing only lines
        # and clear out segments and graphemes here too.
        facsimile.clear_graphemes()
        facsimile.clear_segments()

        ri = tesseract.TessBaseAPIGetIterator(api)
        pi = tesseract.TessResultIteratorGetPageIterator(ri)
        w, h = Image.open(image_path).size
        x0, y0, x1, y1 = (ctypes.c_int(), ctypes.c_int(), ctypes.c_int(),
                          ctypes.c_int())
        i = 0
        lines = facsimile.lines
        while True:
            if tesseract.TessPageIteratorIsAtBeginningOf(pi, RIL_TEXTLINE):
                tesseract.TessPageIteratorBoundingBox(pi, RIL_TEXTLINE,
                                                      ctypes.byref(x0),
                                                      ctypes.byref(y0),
                                                      ctypes.byref(x1),
                                                      ctypes.byref(y1))

                facsimile.scope_line(lines[i][4])
                i += 1
            if tesseract.TessPageIteratorIsAtBeginningOf(pi, RIL_WORD):
                lang = tesseract.TessResultIteratorWordRecognitionLanguage(ri, RIL_WORD).decode('utf-8')
                tesseract.TessPageIteratorBoundingBox(pi, RIL_WORD,
                                                      ctypes.byref(x0),
                                                      ctypes.byref(y0),
                                                      ctypes.byref(x1),
                                                      ctypes.byref(y1))
                conf = tesseract.TessResultIteratorConfidence(ri, RIL_WORD)
                facsimile.add_segment((x0.value, y0.value, x1.value, y1.value),
                                      lang, conf)
            
            conf = tesseract.TessResultIteratorConfidence(ri, RIL_SYMBOL)
            tesseract.TessPageIteratorBoundingBox(pi, RIL_SYMBOL,
                                                  ctypes.byref(x0),
                                                  ctypes.byref(y0),
                                                  ctypes.byref(x1),
                                                  ctypes.byref(y1))
            grapheme = tesseract.TessResultIteratorGetUTF8Text(ri, RIL_SYMBOL).decode('utf-8')
            facsimile.add_graphemes([(grapheme, (x0.value, y0.value, x1.value,
                                      y1.value), conf)])
            if not tesseract.TessResultIteratorNext(ri, RIL_SYMBOL):
                break
        with open(output_path, 'wb') as fp:
            facsimile.write(fp)
        tesseract.TessResultIteratorDelete(ri)
    else:
        with open(output_path, 'wb') as fp:
            tp = tesseract.TessBaseAPIGetHOCRText(api)
            hocr = StringIO.StringIO()
            hocr.write(ctypes.string_at(tp))
            hocr.seek(0)
            facsimile.load_hocr(hocr)
            facsimile.write(fp)
            tesseract.TessDeleteText(tp)
    tesseract.TessBaseAPIDelete(api)


def ocr_direct(image_path, output_path, languages):
    """
    OCRs an image by calling the tesseract executable directly. Images are read
    using the linked leptonica library and the given output_path WILL be
    modified by tesseract.

    Args:
        image_path (unicode): Path to the input image
        output_path (unicode): Path to the hOCR output
        languages (list): List of valid tesseract language identifiers
    """
    p = subprocess.Popen(['tesseract', image_path, output_path, '-l',
                          '+'.join(languages), '-psm', '4', '--tessdata-dir',
                          tessdata, 'hocr'], stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)
    out, err = p.communicate()
    if p.returncode:
        raise NidabaTesseractException(err)




# -*- coding: utf-8 -*-
"""
nidaba.tasks.stats
~~~~~~~~~~~~~~~~~~

Various tasks calculating metrics on documents.

"""

from __future__ import absolute_import, unicode_literals

import os
import difflib

from lxml import html
from pyxdameraulevenshtein import normalized_damerau_levenshtein_distance

from nidaba import storage
from nidaba.tei import TEIFacsimile
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask
from nidaba.algorithms.string import sanitize
from nidaba.nidabaexceptions import NidabaInvalidParameterException


def cleanup(text):
    """
    Removes lines containing only whitespace and normalizes to NFD.
    """
    text = sanitize(text)
    return '\n'.join([s for s in text.splitlines() if len(s.strip())])


@app.task(base=NidabaTask, name=u'nidaba.stats.text_diff_ratio')
def text_diff_ratio(doc, method=u'text_diff_ratio', ground_truth=None,
                    xml_in=True, gt_format=u'tei', clean_in=True, clean_gt=True,
                    divert=True):
    """
    Calculates the similarity of the input documents and a given ground truth
    using the algorithm of python's difflib SequenceMatcher. The result is a
    value between 0.0 (no commonality) and 1.0 (identical strings).

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to the output file.
        ground_truth (unicode): Ground truth location tuple
        xml_in (bool): Switch to treat input as an TEI-XML document.
        gt_format (unicode): Switch to select ground truth format. Valid values
                             are 'tei', 'hocr', and 'text'.
        divert (bool): Switch selecting output diversion. If enabled the output
                       will be added to the tracking arguments and the input
                       document will be returned as the result of the task. Use
                       this to insert a statistical measure into a chain
                       without affecting the results.

    Returns:
        (unicode, unicode): Storage tuple of the output document
    """
    input_path = storage.get_abs_path(*doc[0])
    output_path = storage.insert_suffix(input_path, method,
                                        os.path.basename(input_path))
    with storage.StorageFile(*ground_truth) as fp:
        if gt_format == 'tei':
            tei = TEIFacsimile()
            tei.read(fp)
            gt = '\n'.join(x[-1] for x in tei.lines)
        elif gt_format == 'hocr':
            gt = html.parse(fp).text_content()
        elif gt_format == 'text':
            gt = fp.read()
        else:
            raise NidabaInvalidParameterException('Input format ' + gt_format + ' unknown.')
    with storage.StorageFile(*doc) as fp:
        if xml_in:
            tei = TEIFacsimile()
            tei.read(fp)
            text = '\n'.join(x[-1] for x in tei.lines)
        else:
           text = fp.read()
    if clean_in:
        text = cleanup(text)
    if clean_gt:
        gt = cleanup(gt)
    sm = difflib.SequenceMatcher()
    sm.set_seqs(text, gt)
    if not divert:
        storage.write_text(*get_storage_path(output_path), text=unicode(sm.ratio()))
        return output_path
    else:
        return {'diff_ratio': sm.ratio(), 'doc': doc}


@app.task(base=NidabaTask, name=u'nidaba.stats.text_edit_ratio')
def text_edit_ratio(doc, method=u'text_edit_ratio', ground_truth=None,
                    xml_in=True, gt_format='tei', clean_in=True, clean_gt=True,
                    divert=True):
    """
    Calculates the similarity of the input documents and a given ground truth
    using the Damerau-Levenshtein distance. The result is a value between 0.0
    (no commonality) and 1.0 (identical strings).

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to the output file.
        ground_truth (unicode): Ground truth location tuple
        xml_in (bool): Switch to treat input as an TEI-XML document.
        gt_format (unicode): Switch to select ground truth format. Valid values
                             are 'tei', 'hocr', and 'text'.
        divert (bool): Switch selecting output diversion. If enabled the output
                       will be added to the tracking arguments and the input
                       document will be returned as the result of the task. Use
                       this to insert a statistical measure into a chain
                       without affecting the results.

    Returns:
        (unicode, unicode): Storage tuple of the output document
    """
    input_path = storage.get_abs_path(*doc[0])
    output_path = storage.insert_suffix(input_path, method,
                                        os.path.basename(input_path))
    with storage.StorageFile(*ground_truth) as fp:
        if gt_format == 'tei':
            tei = TEIFacsimile()
            tei.read(fp)
            gt = '\n'.join(x[-1] for x in tei.lines)
        elif gt_format == 'hocr':
            gt = html.parse(fp).text_content()
        elif gt_format == 'text':
            gt = fp.read()
        else:
            raise NidabaInvalidParameterException('Input format ' + gt_format + ' unknown.')
    with storage.StorageFile(*doc) as fp:
        if xml_in:
            tei = TEIFacsimile()
            tei.read(fp)
            text = '\n'.join(x[-1] for x in tei.lines)
        else:
           text = fp.read()
    if clean_in:
        text = cleanup(text)
    if clean_gt:
        gt = cleanup(gt)
    edist = 1.0 - normalized_damerau_levenshtein_distance(text, gt)
    if not divert:
        storage.write_text(*get_storage_path(output_path), text=unicode(edit))
        return output_path
    else:
        return {'edit_ratio': edist, 'doc': doc}



# -*- coding: utf-8 -*-
"""
nidaba.tasks.output
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Various tasks implementing common housekeeping processes like format
conversion, metadata enrichment, ...

"""

from __future__ import absolute_import, unicode_literals

import yaml

from lxml import etree

from nidaba import storage
from nidaba.celery import app
from nidaba.tei import TEIFacsimile
from nidaba.nidabaexceptions import NidabaTEIException
from nidaba.tasks.helper import NidabaTask


@app.task(base=NidabaTask, name=u'nidaba.output.metadata')
def tei_metadata(doc, method=u'metadata', metadata=None, validate=True):
    """
    Enriches a TEI-XML document with various metadata from an user-supplied
    YAML file. 

    The following fields may be contained in the metadata file with the bolded
    subset mandatory for a valid TEI-XML file. They are grouped by their place
    in the header. Unknown fields are ignored and input is escaped as to
    disable 

    Some element may also be extended by increasing their arity, the second
    value is then usually used as a global identifer/locator, i.e. an URL or
    authority control ID.

    titleStmt:

        * ``title``: Title of the resource
        * author: Name of the author of the resource (may be extended)
        * editor: Name of the editor, compiler, translator, etc. of the
                  resource (may be extended)
        * funder: Institution responsible for the funding of the text (may be
                  extended)
        * principal: PI responsible for the creation of the text (may be
                     extended)
        * sponsor: Name of the sponsoring institution (may be extended)
        * meeting: Conference/meeting resulting in the text (may be extended)

    editionStmt:

        * edition: Peculiarities to the underlying edition of the text

    publicationStmt:

        * ``availability``: Licence of the content (may be extended)
        * ``publisher``: Person or agency responsible for the publication of
                     the text (may be extended)
        * distributor: Person or agency responsible for the text's
                       distribution (may be extended)
        * authority: Authority responsible for making the work available
        * idno: Identifier of the publication (may be extended with the type of
                identifier)
        * pubplace: Place of publication
        * date: Date of publication

    seriesStmt:

        * series_title: Title of the series to which the publication belongs

    notesStmt:
        
        * note: Misc. notes about the text

    sourceDesc:

        * ``source_desc``: Description of the source document
   
    other:

        * lang: Abbreviation of the language used in the header

    There is a sample file from the OpenPhilology project in the example
    directory.

    Args:
        doc (unicode, unicode): Storage tuple of the input document
        method (unicode):
        metadata (unicode, unicode): Storage tuple of the metadata YAML file

    Returns:
        (unicode, unicode): Storage tuple of the output document

    Raises:
        NidabaTEIException if the resulting document is not TEI compatible and
        validation is enabled.
    """
    with storage.StorageFile(*doc) as fp:
        tei = TEIFacsimile()
        tei.read(fp)
    with storage.StorageFile(*metadata) as fp:
        meta = yaml.safe_load(fp)
    validFields = ['author', 'editor', 'funder', 'principal', 'sponsor',
                   'meeting', 'edition', 'availability', 'publisher',
                   'distributor', 'authority', 'idno', 'pubplace',
                   'series_title', 'note', 'source_desc', 'lang']
    for field in filter(lambda x: x in validFields, meta):
        setattr(tei, field, meta[field])
    if validate:
        pass
    output_path = storage.insert_suffix(doc[1], method, metadata[1])
    with storage.StorageFile(doc[0], output_path, 'wb') as fp:
        tei.write(fp)
    return (doc[0], output_path)

@app.task(base=NidabaTask, name=u'nidaba.output.tei2simplexml')
def tei2simplexml(doc, method=u'simplexml'):
    """
    Convert a TEI Facsimile to hOCR preserving as much metadata as possible.

    Args:
        doc (unicode, unicode): Storage tuple of the input document

    Returns:
        (unicode, unicode): Storage tuple of the output document
    """
    with storage.StorageFile(*doc) as fp:
        tei = TEIFacsimile()
        tei.read(fp)
    output_path = storage.insert_suffix(doc[1], method)
    with storage.StorageFile(doc[0], output_path, 'wb') as fp:
        tei.write_simplexml(fp)
    return (doc[0], output_path)
   


@app.task(base=NidabaTask, name=u'nidaba.output.tei2hocr')
def tei2hocr(doc, method=u'tei2hocr'):
    """
    Convert a TEI Facsimile to hOCR preserving as much metadata as possible.

    Args:
        doc (unicode, unicode): Storage tuple of the input document

    Returns:
        (unicode, unicode): Storage tuple of the output document
    """
    with storage.StorageFile(*doc) as fp:
        tei = TEIFacsimile()
        tei.read(fp)
    output_path = storage.insert_suffix(doc[1], method)
    with storage.StorageFile(doc[0], output_path, 'wb') as fp:
        tei.write_hocr(fp)
    return (doc[0], output_path)




# -*- coding: utf-8 -*-
"""
nidaba.lex
~~~~~~~~~~

This module contains functions for dealing with words and dictionaries, such as
extracting words from texts, normalizing encodings, building symmetric deletion
dictionaries, etc.
"""

from __future__ import absolute_import

import os
import codecs
import glob
import regex
import nidaba.algorithms.string as alg
from collections import Counter


def tei_spellcheck(facsimile, dictionary, deletion_dictionary,
                   filter_punctuation=False):
    """
    Performs a spell check on an TEI XML document.

    Each ``seg`` element is treated as a single word and spelling corrections
    will be inserted using a choice tag. Correct words will be untouched and
    correction candidates will be sorted by edit distance.

    Args:
        facsimile (nidaba.tei.TEIFacsimile): TEIFacsimile object.
        dictionary (unicode): Path to a base dictionary.
        deletion_dictionary (unicode): Path to a deletion dictionary.
        filter_punctuation (bool): Switch to filter punctuation inside
                                   segments.

    Returns:
        A TEIFacsimile object containing the spelling corrections.
    """
    text_tokens = [x[-1] for x in facsimile.segments]
    if filter_punctuation:
        text_tokens = [regex.sub('[^\w]', '', x) for x in text_tokens]
    suggestions = spellcheck(text_tokens, dictionary, deletion_dictionary)
    facsimile.add_respstmt('spell-checker', 'nidaba-levenshtein')
    for segment in facsimile.segments:
        key = segment[-1]
        if filter_punctuation:
            key = regex.sub('[^\w]', '', key)
        if key not in suggestions:
            continue
        for sugg in suggestions[key]:
            facsimile.add_choices(segment[-2], [(sugg, 100 - 10*alg.edit_distance(key, sugg))])
    return facsimile


def spellcheck(tokens, dictionary, deletion_dictionary):
    """
    Performs a spell check on a sequence of tokens.

    The spelling of each sequence of characters is compared against a
    dictionary containing deletions of valid words and a dictionary of correct
    words.

    Args:
        tokens (iterable): An iterable returning sequences of unicode
                           characters.
        dictionary (unicode): Path to a base dictionary.
        deletion_dictionary (unicode): Path to a deletion dictionary.

    Returns:
        A dictionary containing a sorted (least to highest edit distance) list
        of suggestions for each character sequence that is not contained
        verbatim in the dictionary. Tokens that are not recognized as valid
        words but don't have spelling suggestions either will be contained in
        the result dictionary.
    """
    suggestions = {}
    for tok in tokens:
        tok = alg.sanitize(tok)
        if alg.mmap_bin_search(tok, dictionary,
                           entryparser_fn=alg.key_for_single_word):
            continue
        if tok in suggestions:
            continue
        ret = alg.mapped_sym_suggest(tok, deletion_dictionary, dictionary, 1)
        suggestions[tok] = alg.suggestions(tok, set.union(*ret.itervalues()))
    return suggestions


def cleanlines(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in lines from a file and return them as a sanitized list.
    Non-unique linse will be repeated.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        list: List of lines containing the sanitized output, i.e. normalized
              unicode objects.
    """
    words = []
    with codecs.open(path, u'r', encoding=encoding) as lines:
        for line in lines:
            words.append(alg.sanitize(line, normalization=normalization))
    return words


def cleanwords(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in every word from a files as separated by lines and spaces.
    Non-unique words will be repeated as they are read in. Detects only words
    divided by a standard space.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        list: List of words containing the sanitized output, i.e. normalized
              unicode objects.
    """
    words = []
    with codecs.open(path, u'r', encoding=encoding) as lines:
        for line in lines:
            for seg in line.split(u' '):
                clean = alg.sanitize(seg, normalization=normalization)
                if clean != u'':
                    words.append(clean)
    return words


def uniquewords_with_freq(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in every word from a file as separated by lines and spaces.
    Return a counter (behaves like a dictionary) of unique words
    along with the number of times they occurred.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        Counter: Contains the frequency of each token
    """
    words = cleanwords(path, encoding=encoding, normalization=normalization)
    freq = Counter(words)
    return freq


def cleanuniquewords(path, encoding=u'utf-8', normalization=u'NFD'):
    """
    Read in lines from a file as separated by lines and spaces,
    convert them to the specified normalization, and return a set
    of all unique words.

    Args:
        path (unicode): Absolute path of the file to be read
        encoding (unicode): Encoding to use for decoding the file
        normalization (unicode): Normalization format to use

    Returns:
        set: Set of unique tokens
    """
    return set(cleanwords(path, encoding=encoding,
                          normalization=normalization))


def words_from_files(dirpath, encoding=u'utf-8', normalization=u'NFD'):
    """
    Create a dictionary from a directory of text files.  All files in the given
    directory will be parsed.

    Args:
        dirpath (unicode): Absolute path of the directory to enter
        encoding (unicode): Encoding to use for decoding the files
        normalization (unicode): Normalization format to use

    Returns:
        list: List of words of all files in the directory
    """
    words = []
    for filename in filter(os.path.isfile, glob.glob(dirpath + '/*')):
        words += cleanwords(filename, encoding=encoding,
                            normalization=normalization)
    return words


def unique_words_from_files(dirpath, encoding=u'utf-8', normalization=u'NFD'):
    """
    Create a set of unique words from a directory of text files.  All file in
    the given directory will be parsed.

    Args:
        dirpath (unicode): Absolute path of the directory to enter
        encoding (unicode): Encoding to use for decoding the files
        normalization (unicode): Normalization format to use

    Returns:
        set: Set of words of all files in the directory
    """
    return set(words_from_files(dirpath, encoding=encoding,
                                normalization=normalization))


def make_dict(outpath, iterable, encoding=u'utf-8'):
    """
    Create a file at outpath and write evrey object in iterable to its
    own line. The file is opened in append mode.

    Args:
        outpath (unicode): File path to write to
        iterable (iterable): An iterable used as a data source
        normalization (unicode): Normalization format to use
    """
    with codecs.open(outpath, u'w+', encoding=encoding) as f:
        for s in iterable:
            f.write(s + u'\n')


def make_deldict(outpath, words, depth):
    """
    Creates a symmetric deletion dictionary from the specified word list.

    WARNING! This is a naive approach, which requires all the variants to be
    stored in memory. For large dictionaries at higher depth, this can easily
    use all available memory on most machines.

    Args:
        outpath (unicode): File path to write to
        words (iterable): An iterable returning a single word per iteration
        depth (int): Maximum edit distance to calculate
    """
    variant_dict = {}
    for word in words:
        for var in alg.strings_by_deletion(word, depth):
            if var not in variant_dict:
                variant_dict[var] = []
            variant_dict[var].append(word)
    ordered = sorted(variant_dict.keys())

    with codecs.open(outpath, u'w+', encoding='utf-8') as outfile:
        for key in ordered:
            originals = u' '.join(variant_dict[key])
            outfile.write(u'%s\t%s' % (key, originals) + u'\n')



# -*- coding: utf-8 -*-
"""
nidaba.tei
~~~~~~~~~~

A module for interfacing TEI OCR output
"""

from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals

from lxml import etree
from lxml.etree import Element, SubElement

from collections import OrderedDict
from functools import partial
from copy import deepcopy

from nidaba.nidabaexceptions import NidabaTEIException


def _parse_hocr(title):
    """
    Parses the hOCR title string and returns a dictionary containing its
    contents.
    """
    def int_float_or_str(s):
        try:
            return int(s)
        except ValueError:
            try:
                return float(s)
            except ValueError:
                try:
                    return unicode(s)
                except UnicodeDecodeError:
                    return s

    out = {}
    props = [x.strip() for x in title.split(';')]
    for prop in props:
        p = prop.split()
        out[p[0]] = tuple(int_float_or_str(x) for x in p[1:])
    return out


class TEIFacsimile(object):
    """
    A class encapsulating a TEI XML document following the TEI digital
    facsimile guidelines for embedded transcriptions.
    """

    xml_ns = '{http://www.w3.org/XML/1998/namespace}'
    tei_ns = '{http://www.tei-c.org/ns/1.0}'

    # automatically generated properties and xpath to their location
    fields = {'title': ('.//{0}teiHeader//{0}titleStmt'.format(tei_ns),),
              'author': ('.//{0}teiHeader//{0}titleStmt'.format(tei_ns), 'ref'),
              'editor': ('.//{0}teiHeader//{0}titleStmt'.format(tei_ns), 'ref'),
              'funder': ('.//{0}teiHeader//{0}titleStmt'.format(tei_ns), 'ref'),
              'principal': ('.//{0}teiHeader//{0}titleStmt'.format(tei_ns), 'ref'),
              'sponsor': ('.//{0}teiHeader//{0}titleStmt'.format(tei_ns), 'ref'),
              'meeting': ('.//{0}teiHeader//{0}titleStmt'.format(tei_ns), 'ref'),
              'edition': ('.//{0}teiHeader//{0}editionStmt'.format(tei_ns),),
              'availability': ('.//{0}teiHeader//{0}publicationStmt'.format(tei_ns), 'ref'),
              'publisher': ('.//{0}teiHeader//{0}publicationStmt'.format(tei_ns), 'ref'),
              'distributor': ('.//{0}teiHeader//{0}publicationStmt'.format(tei_ns), 'ref'),
              'authority': ('.//{0}teiHeader//{0}publicationStmt'.format(tei_ns), 'ref'),
              'idno': ('.//{0}teiHeader//{0}publicationStmt'.format(tei_ns), 'type'),
              'pubplace': ('.//{0}teiHeader//{0}publicationStmt'.format(tei_ns),),
              'series_title': ('.//{0}teiHeader//{0}seriesStmt'.format(tei_ns),),
              'note': ('.//{0}teiHeader//{0}notesStmt'.format(tei_ns),),
              'source_desc': ('.//{0}teiHeader//{0}sourceDesc'.format(tei_ns),),
             }

    fileDesc = ['titleStmt', 'editionStmt', 'publicationStmt', 'seriesStmt',
                'notesStmt', 'sourceDesc', ]

    def _generic_getter(self, field):
        el = self.doc.find(self.fields[field][0] + '/{0}{1}'.format(self.tei_ns, field))
        if hasattr(el, 'text'):
            return el.text
        else:
            return None
       
    def _generic_setter(self, value, field):
        el = self.doc.find(self.fields[field][0] + '/{0}{1}'.format(self.tei_ns, field))
        parent = self.doc.find(self.fields[field][0])
        if parent is None:
            _, _, stmt = self.fields[field][0].rpartition('}')
            loc = self.fileDesc.index(stmt)
            while loc:
                loc -= 1
                prev_stmt = self.doc.find('.//{0}{1}'.format(self.tei_ns, self.fileDesc[loc]))
                if prev_stmt is not None:
                    break
            prev_stmt.addnext(Element('{0}{1}'.format(self.tei_ns, stmt)))
        if el is None:
            el = SubElement(self.doc.find(self.fields[field][0]), self.tei_ns + field)
        if isinstance(value, list):
            el.set(self.fields[field][1], value[1])
            value = value[0]
        el.text = value

    def __init__(self):
        doc = Element('TEI', nsmap={None: 'http://www.tei-c.org/ns/1.0'},
                      version='5.0')
        header = SubElement(doc, self.tei_ns + 'teiHeader')
        fileDesc = SubElement(header, self.tei_ns + 'fileDesc')
        SubElement(fileDesc, self.tei_ns + 'titleStmt')
        SubElement(fileDesc, self.tei_ns + 'publicationStmt')

        self.word_scope = None
        self.line_scope = None
        self.resp = None
        self.line_cnt = -1
        self.seg_cnt = -1
        self.grapheme_cnt = -1
        self.doc = doc

    def document(self, dim, image_url):
        sourceDoc = SubElement(self.doc, self.tei_ns + 'sourceDoc')
        surface = SubElement(sourceDoc, self.tei_ns + 'surface', ulx='0',
                             uly='0', lrx=str(dim[0]), lry=str(dim[1]))
        SubElement(surface, self.tei_ns + 'graphic', url=image_url)
        SubElement(surface, self.tei_ns + 'zone')

    @property
    def lang(self):
        """
        The language value of the teiHeader
        """
        el = self.doc.find('.//' + self.tei_ns + 'teiHeader')
        if el is not None:
            return el.get(self.xml_ns + 'lang')
        return None

    @lang.setter
    def lang(self, value):
        el = self.doc.find('.//' + self.tei_ns + 'teiHeader')
        el.set(self.xml_ns + 'lang', value)

    @property
    def description(self):
        """
        Returns a tuple containing a source document's path and its dimensions.
        """
        surface = self.doc.find('//' + self.tei_ns + 'sourceDoc/' + self.tei_ns
                                + 'surface')
        return (surface.get('ulx'),
                surface.get('uly'),
                surface.get('lrx'),
                surface.get('lry'),
                surface.find(self.tei_ns + 'graphic').get('url'))

    @property
    def respstmt(self):
        """
        Returns an ordered dictionary of responsibility statements from the XML
        document.
        """
        d = OrderedDict()
        for resp in self.doc.iter(self.tei_ns + 'respStmt'):
            rip = resp.get(self.xml_ns + 'id')
            d[rip] = {c.tag: c.text for c in resp.getchildren()}
        return d

    def add_respstmt(self, name, resp):
        """
        Adds a responsibility statement and treats all subsequently added text
        as a responsibility of this statement.

        Args:
            name (unicode): Identifier of the process that generated the
                            output.
            resp (unicode): Phrase describing the nature of the process
                            generating the output.

        Returns:
            A unicode string corresponding to the responsibility identifier.
        """
        id = -1
        for rstmt in self.doc.iter(self.tei_ns + 'respStmt'):
            id += 1
        r = SubElement(self.doc.find('.//' + self.tei_ns + 'titleStmt'),
                       self.tei_ns + 'respStmt')
        r.set(self.xml_ns + 'id', u'resp_' + unicode(id + 1))
        self.resp = r.get(self.xml_ns + 'id')
        SubElement(r, self.tei_ns + 'resp').text = resp
        SubElement(r, self.tei_ns + 'name').text = name
        return r.get(self.xml_ns + 'id')

    def scope_respstmt(self, id):
        """
        Scopes a respStmt for subsequent addition of graphemes/segments.

        Args:
            id (unicode): XML id of the responsibility statement

        Raises:
            NidabaTEIException if the identifier is unknown
        """
        if self.doc.find(".//" + self.tei_ns + "respStmt[@" + self.xml_ns +
                         "id='" + id + "']") is None:
            raise NidabaTEIException('No such responsibility statement.')
        self.resp = id

    @property
    def lines(self):
        """
        Returns an reading order sorted list of tuples in the format (x0, y0,
        x1, y1, xml id, text).
        """
        lines = []
        for line in self.doc.iter(self.tei_ns + 'line'):
            text = ''.join(line.itertext())
            lines.append((int(line.get('ulx')), int(line.get('uly')),
                          int(line.get('lrx')), int(line.get('lry')),
                          line.get(self.xml_ns + 'id'), text))
        return lines

    def add_line(self, dim):
        """
        Marks the beginning of a new topographical line and scopes it.

        Args:
            dim (tuple): A tuple containing the bounding box (x0, y0, x1, y1)
        """
        zone = self.doc.find('.//' + self.tei_ns + 'zone')
        self.line_scope = SubElement(zone, self.tei_ns + 'line',
                                     ulx=str(dim[0]), uly=str(dim[1]),
                                     lrx=str(dim[2]), lry=str(dim[3]))
        self.line_cnt += 1
        self.line_scope.set(self.xml_ns + 'id', 'line_' + str(self.line_cnt))
        if self.resp:
            self.line_scope.set('resp', '#' + self.resp)
        self.word_scope = None

    def scope_line(self, id):
        """
        Scopes a particular line for addition of segments/graphemes. Also
        disables the current segment scope.

        Args:
            id (unicode): XML id of the line tag

        Raises:
            NidabaTEIException if the identifier is unknown
        """
        line = self.doc.find(".//" + self.tei_ns + "line[@" + self.xml_ns +
                             "id='" + id + "']")
        if line is None:
            raise NidabaTEIException('No such line')
        self.line_scope = line

    @property
    def segments(self):
        """
        Returns an reading order sorted list of tuples in the format (x0, y0,
        x1, y1, confidence, id, text).
        """
        segments = []
        for seg in self.doc.iter(self.tei_ns + 'seg'):
            text = ''.join(seg.itertext())
            if seg.getparent().get('type') == 'word':
                bbox = (int(seg.getparent().get('ulx')),
                        int(seg.getparent().get('uly')),
                        int(seg.getparent().get('lrx')),
                        int(seg.getparent().get('lry')))
            else:
                bbox = (None, None, None, None)
            cert = self.doc.xpath("//*[local-name()='certainty' and @target=$tag]",
                                  tag = '#' + seg.get(self.xml_ns + 'id'))
            if len(cert):
                cert = int(100.0 * float(cert[0].get('degree')))
            else:
                cert = None
            segments.append(bbox + (cert,) + (seg.get(self.xml_ns + 'id'), text))
        return segments

    def add_segment(self, dim, lang=None, confidence=None):
        """
        Marks the beginning of a new topographical segment in the current
        scope. Most often this correspond to a word recognized by an engine.

        Args:
            dim (tuple): A tuple containing the bounding box (x0, y0, x1, y1)
            lang (unicode): Optional identifier of the segment language.
            confidence (float): Optional confidence value between 0 and 100.
        """
        zone = SubElement(self.line_scope, self.tei_ns + 'zone', 
                          ulx=str(dim[0]), uly=str(dim[1]), lrx=str(dim[2]),
                          lry=str(dim[3]), type='word')
        self.word_scope = SubElement(zone, self.tei_ns + 'seg')
        self.seg_cnt += 1
        self.word_scope.set(self.xml_ns + 'id', 'seg_' + str(self.seg_cnt))
        if confidence:
            cert = SubElement(self.word_scope, self.tei_ns + 'certainty',
                              degree = u'{0:.2f}'.format(confidence/100.0),
                              locus = 'value',
                              target = '#' + 'seg_' + str(self.seg_cnt))
            if self.resp:
                cert.set('resp', '#' + self.resp)
        if self.resp:
            self.word_scope.set('resp', '#' + self.resp)

    def clear_segment(self):
        """
        Marks the end of the current topographical segment.
        """
        self.word_scope = None

    @property
    def graphemes(self):
        """
        Returns a reading order sorted list of tuples in the format (x0, y0,
        x1, y1, id, text).
        """
        graphemes = []
        for g in self.doc.iter(self.tei_ns + 'g'):
            text = ''.join(g.itertext())
            if g.getparent().get('type') == 'grapheme':
                bbox = (int(g.getparent().get('ulx')),
                        int(g.getparent().get('uly')),
                        int(g.getparent().get('lrx')),
                        int(g.getparent().get('lry')))
            else:
                bbox = (None, None, None, None)
            cert = self.doc.xpath("//*[local-name()='certainty' and @target=$tag]",
                                  tag = '#' + g.get(self.xml_ns + 'id'))
            if len(cert):
                cert = int(100.0 * float(cert[0].get('degree')))
            else:
                cert = None
            graphemes.append(bbox + (cert,) + (g.get(self.xml_ns + 'id'), text))
        return graphemes

    def add_graphemes(self, it):
        """
        Adds a number of graphemes to the current scope (either line or word).
        A line or segment has to be created beforehand.

        Args:
            it (iterable): An iterable returning a tuple containing a glyph
                           (unicode), and optionally the bounding box of this
                           glyph (x0, y0, x1, y1) and a recognition confidence
                           value in the range 0 and 100.
        """
        scope = self.word_scope if self.word_scope is not None else self.line_scope
        for t in it:
            conf = None
            if len(t) == 1:
                g = t
                zone = scope
            else:
                if len(t) == 2:
                    g, box = t
                else:
                    g, box, conf = t
                ulx, uly, lrx, lry = box
                zone = SubElement(scope, self.tei_ns + 'zone', ulx=str(ulx),
                                  uly=str(uly), lrx=str(lrx), lry=str(lry),
                                  type='grapheme', resp= '#' + self.resp)
            glyph = SubElement(zone, self.tei_ns + 'g')
            self.grapheme_cnt += 1
            glyph.set(self.xml_ns + 'id', 'grapheme_' + str(self.grapheme_cnt))
            glyph.text = g
            if conf:
                cert = SubElement(zone, self.tei_ns + 'certainty',
                                  degree=u'{0:.2f}'.format(conf/100.0),
                                  locus = 'value',
                                  target = '#' + 'grapheme_' +
                                  str(self.grapheme_cnt))
                if self.resp:
                    cert.set('resp', '#' + self.resp)
            if self.resp:
                glyph.set('resp', '#' + self.resp)
    
    def add_choices(self, id, it):
        """
        Adds alternative interpretations to an element.

        Args:
            id (unicode): 
            it (iterable): An iterable returning a tuple containing an
                           alternative reading and an optional confidence value
                           in the range between 0 and 100.
        """
        el = self.doc.xpath("//*[@xml:id=$tagid]", tagid = id)[0]
        # remove old tree only if not already part of an choice segment.
        parent = el.getparent()
        if parent.tag == self.tei_ns + 'sic':
            choice = parent.find('..')
        else:
            sic = deepcopy(el)
            parent.remove(el)
            choice = SubElement(parent, self.tei_ns + 'choice')
            # reinsert beneath sic element
            SubElement(choice, self.tei_ns + 'sic').append(sic)
        for alt in it:
            corr = SubElement(choice, self.tei_ns + 'corr')
            if self.resp:
                corr.set('resp', '#' + self.resp)
            corr.text = alt[0]
            if len(alt) == 2:
                cert = SubElement(corr, self.tei_ns + 'certainty',
                                  degree = u'{0:.2f}'.format(alt[1]/100.0),
                                  locus = 'value')
                if self.resp:
                    cert.set('resp', '#' + self.resp)

    def clear_lines(self):
        """
        Deletes all <line> nodes from the document.
        """
        for zone in self.doc.iterfind('.//' + self.tei_ns +
                                      "line"):
            zone.getparent().remove(zone)
        self.line_scope = None
        self.word_scope = None
        self.line_cnt = -1
        self.seg_cnt = -1
        self.grapheme_cnt = -1

    def clear_graphemes(self):
        """
        Deletes all <g> nodes from the document. Mainly used when combining
        page segmentation algorithms extracting graphemes and OCR engines
        operating on lexemes. Also resets the current scope to the first line
        (and if applicable its first segment).
        """
        for zone in self.doc.iterfind('.//' + self.tei_ns +
                                      "zone[@type='grapheme']"):
            zone.getparent().remove(zone)
        self.line_scope = self.doc.find('.//' + self.tei_ns + 'line')
        self.word_scope = self.doc.find('.//' + self.tei_ns + 'seg')
        self.grapheme_cnt = -1

    def clear_segments(self):
        """
        Deletes all <seg> nodes from the document. Mainly used when combining
        page segmentation algorithms extracting lexemes (and graphemes) and OCR
        engines operating on lines. Also resets the current scope to the first
        line.
        """
        for zone in self.doc.iterfind('.//' + self.tei_ns +
                                      "zone[@type='word']"):
            zone.getparent().remove(zone)
        self.line_scope = self.doc.find('.//' + self.tei_ns + 'line')
        self.word_scope = None
        self.seg_cnt = -1
        self.grapheme_cnt = -1

    def load_hocr(self, fp):
        """
        Extracts as much information as possible from an hOCR file and converts
        it to TEI.

        TODO: Write a robust XSL transformation.

        Args:
            fp (file): File descriptor to read data from.
        """
        doc = etree.HTML(fp.read())
        self.clear_lines()
        self.clear_segments()
        self.clear_graphemes()
        el = doc.find(".//meta[@name='ocr-system']")
        if el is not None:
            self.add_respstmt(el.get('content'), 'ocr-system')
        page = doc.find('.//div[@class="ocr_page"]')
        o = _parse_hocr(page.get('title'))
        self.document(o['bbox'], o['image'][0])
        for line in doc.iterfind('.//span[@class="ocr_line"]'):
            self.add_line(_parse_hocr(line.get('title'))['bbox'])
            if not line.xpath('.//span[starts-with(@class, "ocrx")]'):
                self.add_graphemes(''.join(line.itertext()))
            for span in line.xpath('.//span[starts-with(@class, "ocrx")]'):
                o = _parse_hocr(span.get('title'))
                confidence = None
                bbox = None
                if 'bbox' in o:
                    bbox = o['bbox']
                if 'x_wconf' in o:
                    confidence = int(o['x_wconf'][0])
                self.add_segment(bbox, confidence=confidence)
                self.add_graphemes(''.join(span.itertext()))
                if span.tail:
                    self.clear_segment()
                    # strip trailing whitespace as some engines add it
                    # arbitrarily or for formatting purposes
                    self.add_graphemes(span.tail.rstrip())

    def write_hocr(self, fp):
        """
        Writes the TEI document as an hOCR file.

        Args:
            fp (file): File descriptor to write to.
        """
        page = etree.Element('html', xmlns="http://www.w3.org/1999/xhtml")
        head = SubElement(page, 'head')
        SubElement(head, 'title').text = self.title
        SubElement(head, 'meta', name="ocr-system",
                   content=self.respstmt.values()[-1][self.tei_ns + 'name'])
        capa = "ocr_page"
        if self.lines is not None:
            capa += ", ocr_line"
        if self.segments is not None:
            capa += ", ocrx_word"
        SubElement(head, 'meta', name='ocr-capabilities', content=capa)
        body = SubElement(page, 'body')
        ocr_page = SubElement(body, 'div', title='')
        ocr_page.set('class', 'ocr_page')
        for line in self.doc.iter(self.tei_ns + 'line'):
            ocr_line = SubElement(ocr_page, 'span')
            ocr_line.set('class', 'ocr_line')
            ocr_line.set('title', 'bbox ' + ' '.join([str(line.get('ulx')),
                                                      str(line.get('uly')),
                                                      str(line.get('lrx')),
                                                      str(line.get('lry'))]))
            # get text not in word segments interleaved with segments
            for seg in line.xpath('child::node()'):
                if isinstance(seg, etree._ElementStringResult):
                    if ocr_line.text is None:
                        ocr_line.text = ''
                    ocr_line.text += seg
                else:
                    # zone from 
                    ocrx_word = SubElement(ocr_line, 'span')
                    ocrx_word.set('class', 'ocrx_word')
                    title = 'bbox ' + ' '.join([str(seg.get('ulx')),
                                               str(seg.get('uly')),
                                               str(seg.get('lrx')),
                                               str(seg.get('lry'))])
                    cert = seg.find('.//{0}certainty'.format(self.tei_ns))
                    if cert is not None:
                        title += '; x_wconf ' + str(int(100.0 *
                                                    float(cert.get('degree'))))
                    ocrx_word.set('title', title)
                    ocrx_word.text = ''.join(seg.itertext())
            SubElement(ocr_page, 'br')
        fp.write(etree.tostring(page, pretty_print=True, 
                 doctype='<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 '
                 'Transitional//EN" '
                 '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">',
                 xml_declaration=True, encoding='utf-8'))

    def write_simplexml(self, fp):
        """
        Writes the TEI document as a grapheme cloud in a simple XML format. Its basic format is:

        <text>
        <charParams l="0" r="78" t="6" b="89" charConfidence="76">D</charParams>
        <charParams l="86" r="111" t="24" b="89" charConfidence="76">e</charParams>
        ....
        </text>

        Args:
            fp (file): File descriptor to write to.
        """
        page = etree.Element('text')
        for g in self.graphemes:
            el = SubElement(page, 'charParams')
            el.text = g[-1]
            if g[0] is not None:
                el.set('l', str(g[0]))
                el.set('t', str(g[1]))
                el.set('r', str(g[2]))
                el.set('b', str(g[3]))
            if g[-3] is not None:
                el.set('charConfidence', str(g[-3]))
        fp.write(etree.tostring(page, pretty_print=True,
                 xml_declaration=True, encoding='utf-8'))

    def write(self, fp):
        """
        Writes the TEI XML document to a file object.

        Args:
            fp (file): file object to write to
        """
        fp.write(etree.tostring(self.doc, pretty_print=True,
                                xml_declaration=True, encoding='utf-8'))

    def read(self, fp):
        """
        Reads an XML document from a file object and populates all recognized
        attributes. Also sets the scope to the first line (and if applicable
        segment) of the document.

        Args:
            fp (file): file object to read from
        """
        self.doc = etree.parse(fp).getroot()
        self.line_cnt = len(list(self.doc.iter(self.tei_ns + 'line'))) - 1
        self.seg_cnt = len(list(self.doc.iter(self.tei_ns + 'seg'))) - 1
        self.grapheme_cnt = len(list(self.doc.iter(self.tei_ns + 'g'))) - 1


# populate properties on the class. Note that it operates on the TEIFacsimile
# class itself.
for field in TEIFacsimile.fields:
    setattr(TEIFacsimile, field, property(partial(TEIFacsimile._generic_getter,
            field=field), partial(TEIFacsimile._generic_setter, field=field)))



# -*- coding: utf-8 -*-
import unittest
import os
import tempfile
import numpy
import mmap

from mock import patch, MagicMock

thisfile = os.path.abspath(os.path.dirname(__file__))
resources = os.path.abspath(os.path.join(thisfile, 'resources/tesseract'))


class LevenshteinTests(unittest.TestCase):
    """Tests the edit_distance funtion by checking both the edit distance and
    full matrix comparisons."""

    # -------------------------------------------------------------------
    # Edit distance tests -----------------------------------------------
    # -------------------------------------------------------------------

    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions


    def test_equal_strings(self):
        """
        Test the edit distance of identical strings
        """
        self.assertEqual(
            0, self.string.edit_distance('test_string', 'test_string'))

    def test_single_insert(self):
        """
        Test with strings requiring one insert
        """
        self.assertEqual(1, self.string.edit_distance('', 'a'))

    def test_tenfold_insert(self):
        """
        Test with strings requiring ten inserts
        """
        self.assertEqual(10, self.string.edit_distance('', 'aaaaaaaaaa'))

    def test_single_delete(self):
        """
        Test with strings requiring one delete
        """
        self.assertEqual(1, self.string.edit_distance('a', ''))

    def test_tenfold_delete(self):
        """
        Test with strings requiring ten deletes
        """
        self.assertEqual(10, self.string.edit_distance('aaaaaaaaaa', ''))

    def test_singledelete_singleadd(self):
        """
        Test with strings requiring a single non-trivial insert.
        """
        # Should delete the a and insert the final b.
        self.assertEqual(2, self.string.edit_distance('abbb', 'bbbbb'))

    def test_singleadd_singledelete(self):
        """
        Test with strings requiring a single non-trivial delete.
        """

        # Should delete the a and add the final b.
        self.assertEqual(2, self.string.edit_distance('bbbbb', 'abbb'))

    # -------------------------------------------------------------------
    # Score matrix tests ------------------------------------------------
    # -------------------------------------------------------------------

    def test_match_delete_matrix(self):
        """
        Test the scoring matrix state in an edit requiring one match and
        delete.
        """
        expected = [[0, 1, 2], [1, 0, 1]]
        self.assertEqual(expected,
                         self.string.native_full_edit_distance('a', 'ab')[0])

    def test_match_insert_matrix(self):
        """
        Test the scoring matrix state in an edit requiring one match an insert.
        """
        expected = [[0, 1], [1, 0], [2, 1]]
        self.assertEqual(
            expected, self.string.native_full_edit_distance('ab', 'a')[0])

    def test_all_match_matrix(self):
        """
        Test the scoring matrix state in an edit consisting only of multiple
        matches.
        """
        expected = [[0, 1, 2, 3, 4, 5],
                    [1, 0, 1, 2, 3, 4],
                    [2, 1, 0, 1, 2, 3],
                    [3, 2, 1, 0, 1, 2],
                    [4, 3, 2, 1, 0, 1],
                    [5, 4, 3, 2, 1, 0]]
        self.assertEqual(expected,
                         self.string.native_full_edit_distance('aaaaa',
                                                              'aaaaa')[0])

    def test_all_subtitute_matrix(self):
        """
        Test the scoring matrix state in an edit consisting only of multiple
        substitutes.
        """
        expected = [[0, 1, 2, 3, 4, 5],
                    [1, 1, 2, 3, 4, 5],
                    [2, 2, 2, 3, 4, 5],
                    [3, 3, 3, 3, 4, 5],
                    [4, 4, 4, 4, 4, 5],
                    [5, 5, 5, 5, 5, 5]]
        self.assertEqual(expected,
                         self.string.native_full_edit_distance('aaaaa',
                                                              'bbbbb')[0])

    def test_all_insert(self):
        """
        Test the scoring matrix state in an edit consisting only of multiple
        inserts.
        """
        expected = [[0, 1, 2, 3, 4, 5]]
        self.assertEqual(expected,
                         self.string.native_full_edit_distance('', 'abcde')[0])

    def test_all_delete(self):
        """
        Test the scoring matrix state in an edit consisting only of deletes.
        """
        expected = [[0],
                    [1],
                    [2],
                    [3],
                    [4],
                    [5]]
        self.assertEqual(expected,
                         self.string.native_full_edit_distance('abcde', '')[0])

    def test_all_empty(self):
        """
        Test the scoring matrix state in an edit between two empty strings.
        """
        expected = [[0]]
        self.assertEqual(expected,
                         self.string.native_full_edit_distance('', '')[0])

    # -------------------------------------------------------------------
    # Parameter tests ---------------------------------------------------
    # -------------------------------------------------------------------

    def test_substitution_parameter(self):
        """
        Test the functionality of the substitution score parameter.
        """
        default = self.string.edit_distance('a', 'b')
        same_as_default = self.string.edit_distance(
            'a', 'b', substitutionscore=1)
        with_parameter = self.string.edit_distance(
            'a', 'b', substitutionscore=-7)
        with_parameter_and_max = self.string.edit_distance(
            'a', 'b', substitutionscore=7)
        self.assertEqual(1, default)
        self.assertEqual(1, same_as_default)
        self.assertEqual(-7, with_parameter)

    def test_deletion_parameter(self):
        """
        Test the functionality of the delete score parameter.
        """
        default = self.string.edit_distance('a', '')
        same_as_default = self.string.edit_distance('a', '', deletescore=1)
        with_parameter = self.string.edit_distance('a', '', deletescore=-7)

        self.assertEqual(1, default)
        self.assertEqual(1, same_as_default)
        self.assertEqual(-7, with_parameter)

    def test_insertion_parameter(self):
        """
        Test the functionality of the insert score parameter.
        """
        default = self.string.edit_distance('', 'a')
        same_as_default = self.string.edit_distance('', 'a', insertscore=1)
        with_parameter = self.string.edit_distance('', 'a', insertscore=-7)

        self.assertEqual(1, default)
        self.assertEqual(1, same_as_default)
        self.assertEqual(-7, with_parameter)

    def test_charmatrix_one_character_substitution_lower(self):
        """
        Test the functionality of the charmatrix parameter where in reduces the
        score of a substitution.
        """
        charmatrix = {('a', 'b'): 0}
        expected = [[0, 1],
                    [1, 0]]
        self.assertEqual(expected, self.string.native_full_edit_distance('a',
                                                                        'b',
                                                                        charmatrix=charmatrix)[0])

    def test_charmatrix_one_character_substitution_higher(self):
        """
        Test the functionality of the charmatrix parameter where in increases
        the score of a substitution.
        """
        charmatrix = {('a', 'b'): 5}
        expected = [[0, 1],
                    [1, 5]]
        self.assertEqual(expected, self.string.native_full_edit_distance(
            'a', 'b', charmatrix=charmatrix)[0])

    def test_charmatrix_one_character_substitution_default(self):
        """
        Test the functionality of the charmatrix parameter where it does not
        change a character score.
        """
        charmatrix = {('a', 'b'): 1}
        expected = [[0, 1],
                    [1, 1]]
        self.assertEqual(expected, self.string.native_full_edit_distance(
            'a', 'b', charmatrix=charmatrix)[0])

    def test_charmatrix_default_delete(self):
        """
        Test the case where the charmatrix mutates the score of trivial (first
        column) deletes.
        """
        charmatrix = {('', 'a'): 5}
        expected = [[0],
                    [5],
                    [10],
                    [15]]
        self.assertEqual(expected, self.string.native_full_edit_distance(
            'aaa', '', charmatrix=charmatrix)[0])

    def test_charmatrix_default_insert(self):
        """
        Test the case where the charmatrix mutates the score of trivial (first
        row) inserts.
        """
        charmatrix = {('a', ''): 5}
        expected = [[0, 5, 10, 15]]
        self.assertEqual(expected, self.string.native_full_edit_distance(
            '', 'aaa', charmatrix=charmatrix)[0])
    # -------------------------------------------------------------------
    # Word tests --------------------------------------------------------
    # -------------------------------------------------------------------

    def test_identical_word(self):
        self.assertEqual(0, self.string.edit_distance(['word'], ['word']))

    def test_word_substitution(self):
        self.assertEqual(
            1, self.string.edit_distance(['word'], ['different word']))

    def test_word_insertion(self):
        self.assertEqual(1, self.string.edit_distance([], ['word']))

    def test_word_deletion(self):
        self.assertEqual(1, self.string.edit_distance(['word'], []))

    def test_word_multi_insert(self):
        self.assertEqual(5, self.string.edit_distance([], ['word', 'word',
                                                          'word', 'word',
                                                          'word']))

    def test_word_multi_delete(self):
        self.assertEqual(5, self.string.edit_distance(['word', 'word', 'word',
                                                      'word', 'word'], []))

    def test_word_multi_match(self):
        self.assertEqual(0, self.string.edit_distance(['word1', 'word2',
                                                      'word3'], ['word1',
                                                                 'word2',
                                                                 'word3']))

    def test_word_multi_substitute(self):
        self.assertEqual(3, self.string.edit_distance(['word1', 'word2',
                                                      'word3'], ['otherword1',
                                                                 'otherword2',
                                                                 'otherword3']))


class AlignmentTests(unittest.TestCase):

    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions

    def test_empty_strings(self):
        """
        Tests the edit steps between two empty strings.
        """
        self.assertEqual([], self.string.native_align('', ''))

    def test_insertions(self):
        """
        Test the edit steps when only inserts are needed.
        """
        self.assertEqual(['i', 'i', 'i', 'i', 'i'],
                         self.string.native_align('', 'abcde'))

    def test_deletions(self):
        """
        Test the edit steps when only deletes are needed.
        """
        self.assertEqual(['d', 'd', 'd', 'd', 'd'],
                         self.string.native_align('abcde', ''))

    def test_matches(self):
        """
        Test the edit steps when only matches are needed.
        """
        self.assertEqual(['m', 'm', 'm', 'm', 'm'],
                         self.string.native_align('abcde', 'abcde'))

    def test_matches_1(self):  # TODO
        """
        Test the edit steps when only substitutions are needed.
        """
        self.assertEqual(['s', 's', 's', 's', 's'],
                         self.string.native_align('abcde', 'vwxyz'))

    def test_wikipedia_example_1(self):
        """
        Test against the first general example provided at
        http://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm
        """
        self.assertEqual(['s', 'm', 'm', 'm', 's', 'm', 'd'],
                         self.string.native_align('sitting', 'kitten'))

    def test_wikipedia_example_2(self):
        """
        Test against the second general example provided at
        http://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm
        """
        self.assertEqual(['m', 'i', 'i', 'm', 's', 'm', 'm', 'm'],
                         self.string.native_align('sunday', 'saturday'))


class SemiGlobalAlignmentTests(unittest.TestCase):

    """
    Tests the semi_global_align function.
    """
    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions

    def test_sequence_length_inversion(self):
        """
        Test that an exception is thrown if the first sequence is > the second.
        """
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.native_semi_global_align, 'abcdefghi', '')
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.native_semi_global_align, 'a', '')
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.native_semi_global_align, ['word'], [])
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.native_semi_global_align, ['word1',
                                                                'word2',
                                                                'word3'],
                          ['word'])

    def test_identical(self):
        """
        Test with strings and lists of equal length.
        """
        expected = ['m']
        self.assertEqual(
            expected, self.string.native_semi_global_align('a', 'a'))
        self.assertEqual(
            expected, self.string.native_semi_global_align(['a'], ['a']))

    def test_identical_long(self):
        """
        Test with a longer series of matches.
        """
        expected = ['m', 'm', 'm', 'm', 'm']
        self.assertEqual(
            expected, self.string.native_semi_global_align('abcde', 'abcde'))

    def test_identical_words(self):
        """
        Test with a series of of matching words.
        """
        expected = ['m', 'm', 'm', 'm', 'm']
        self.assertEqual(expected, self.string.native_semi_global_align(
            ['word1', 'word2', 'word3', 'word4', 'word5'], ['word1', 'word2',
                                                            'word3', 'word4',
                                                            'word5']))

    def test_simple_prefix(self):
        """
        Test with a single match and a skippable prefix.
        """
        expected = ['i', 'i', 'i', 'i', 'm']
        self.assertEqual(
            expected, self.string.native_semi_global_align('b', 'aaaab'))

    def test_prefix_with_trailer(self):
        """
        Test with a single match and a skippable prefix and trailer.
        """
        expected = ['i', 'i', 'i', 'i', 'm']
        self.assertEqual(
            expected, self.string.native_semi_global_align('b', 'aaaabcccc'))

    def test_word_prefix(self):
        """
        Test a list of words with a prefix.
        """
        expected = ['i', 'i', 'm']
        self.assertEqual(expected, self.string.native_semi_global_align(
            ['match'], ['w1', 'w2', 'match']))


class NumpyLevenshteinTests(unittest.TestCase):

    """
    Tests the np_full_edit_distance funtion by checking both the edit distance
    and full matrix comparisons.
    """
    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions

    # -------------------------------------------------------------------
    # Edit distance tests -----------------------------------------------
    # -------------------------------------------------------------------

    def test_equal_strings(self):
        """
        Test the edit distance of identical strings
        """
        self.assertEqual(0, self.string.np_full_edit_distance(
            'test_string', 'test_string')[0][-1, -1])

    def test_single_insert(self):
        """
        Test with strings requiring one insert
        """
        self.assertEqual(1, self.string.np_full_edit_distance('', 'a')[0][-1,
                                                                         -1])

    def test_tenfold_insert(self):
        """
        Test with strings requiring ten inserts
        """
        self.assertEqual(10, self.string.np_full_edit_distance(
            '', 'aaaaaaaaaa')[0][-1, -1])  # A string with 10 a's.

    def test_single_delete(self):
        """
        Test with strings requiring one delete
        """
        self.assertEqual(
            1, self.string.np_full_edit_distance('a', '')[0][-1, -1])

    def test_tenfold_delete(self):
        """
        Test with strings requiring ten deletes
        """
        self.assertEqual(10, self.string.np_full_edit_distance(
            'aaaaaaaaaa', '')[0][-1, -1])  # A string with 10 a's.

    def test_singledelete_singleadd(self):
        """
        Test with strings requiring a single non-trivial insert.
        """
        self.assertEqual(2, self.string.np_full_edit_distance('abbb',
                                                             'bbbbb')[0][-1,
                                                                         -1])

    def test_singleadd_singledelete(self):
        """
        Test with strings requiring a single non-trivial delete.
        """
        self.assertEqual(2, self.string.np_full_edit_distance('bbbbb',
                                                             'abbb')[0][-1,
                                                                        -1])
    # -------------------------------------------------------------------
    # Score matrix tests ------------------------------------------------
    # -------------------------------------------------------------------

    def test_match_delete_matrix(self):
        """
        Test the scoring matrix state in an edit requiring one match and
        delete.
        """
        expected = numpy.array([[0, 1, 2], [1, 0, 1]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('a',
                                                                           'ab')[0]))

    def test_match_insert_matrix(self):
        """
        Test the scoring matrix state in an edit requiring one match an insert.
        """
        expected = numpy.array([[0, 1], [1, 0], [2, 1]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('ab',
                                                                           'a')[0]))

    def test_all_match_matrix(self):
        """
        Test the scoring matrix state in an edit consisting only of multiple
        matches.
        """
        expected = numpy.array([[0, 1, 2, 3, 4, 5],
                                [1, 0, 1, 2, 3, 4],
                                [2, 1, 0, 1, 2, 3],
                                [3, 2, 1, 0, 1, 2],
                                [4, 3, 2, 1, 0, 1],
                                [5, 4, 3, 2, 1, 0]])
        self.assertTrue(numpy.array_equal(
            expected, self.string.np_full_edit_distance('aaaaa', 'aaaaa')[0]))

    def test_all_subtitute_matrix(self):
        """
        Test the scoring matrix state in an edit consisting only of multiple
        substitutes.
        """
        expected = numpy.array([[0, 1, 2, 3, 4, 5],
                                [1, 1, 2, 3, 4, 5],
                                [2, 2, 2, 3, 4, 5],
                                [3, 3, 3, 3, 4, 5],
                                [4, 4, 4, 4, 4, 5],
                                [5, 5, 5, 5, 5, 5]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('aaaaa',
                                                                           'bbbbb')[0]))

    def test_all_insert(self):
        """
        Test the scoring matrix state in an edit consisting only of multiple
        inserts.
        """
        expected = numpy.array([[0, 1, 2, 3, 4, 5]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('',
                                                                           'abcde')[0]))

    def test_all_delete(self):
        """
        Test the scoring matrix state in an edit consisting only of deletes.
        """
        expected = numpy.array([[0],
                                [1],
                                [2],
                                [3],
                                [4],
                                [5]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('abcde',
                                                                           '')[0]))

    def test_all_empty(self):
        """
        Test the scoring matrix state in an edit between two empty strings.
        """
        expected = numpy.zeros(shape=(1, 1))
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('',
                                                                           '')[0]))
    # -------------------------------------------------------------------
    # Parameter tests ---------------------------------------------------
    # -------------------------------------------------------------------

    def test_substitution_parameter(self):
        """
        Test the functionality of the substitution score parameter.
        """
        default = self.string.np_full_edit_distance('a', 'b')[0][-1, -1]
        same_as_default = self.string.np_full_edit_distance(
            'a', 'b', substitutionscore=1)[0][-1, -1]
        with_parameter = self.string.np_full_edit_distance(
            'a', 'b', substitutionscore=-7)[0][-1, -1]
        with_parameter_and_max = self.string.np_full_edit_distance(
            'a', 'b', substitutionscore=7)[0][-1, -1]
        self.assertEqual(1, default)
        self.assertEqual(1, same_as_default)
        self.assertEqual(-7, with_parameter)

    def test_deletion_parameter(self):
        """
        Test the functionality of the delete score parameter.
        """
        default = self.string.np_full_edit_distance('a', '')[0][-1, -1]
        same_as_default = self.string.np_full_edit_distance(
            'a', '', deletescore=1)[0][-1, -1]
        with_parameter = self.string.np_full_edit_distance(
            'a', '', deletescore=-7)[0][-1, -1]

        self.assertEqual(1, default)
        self.assertEqual(1, same_as_default)
        self.assertEqual(-7, with_parameter)

    def test_insertion_parameter(self):
        """
        Test the functionality of the insert score parameter.
        """
        default = self.string.np_full_edit_distance('', 'a')[0][-1, -1]
        same_as_default = self.string.np_full_edit_distance(
            '', 'a', insertscore=1)[0][-1, -1]
        with_parameter = self.string.np_full_edit_distance(
            '', 'a', insertscore=-7)[0][-1, -1]

        self.assertEqual(1, default)
        self.assertEqual(1, same_as_default)
        self.assertEqual(-7, with_parameter)

    def test_charmatrix_one_character_substitution_lower(self):
        """
        Test the functionality of the charmatrix parameter where in reduces the
        score of a substitution.
        """
        charmatrix = {('a', 'b'): 0}
        expected = numpy.array([[0, 1],
                                [1, 0]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('a',
                                                                           'b',
                                                                           charmatrix=charmatrix)[0]))

    def test_charmatrix_one_character_substitution_higher(self):
        """
        Test the functionality of the charmatrix parameter where in increases
        the score of a substitution.
        """
        charmatrix = {('a', 'b'): 5}
        expected = numpy.array([[0, 1],
                                [1, 5]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('a',
                                                                           'b',
                                                                           charmatrix=charmatrix)[0]))

    def test_charmatrix_one_character_substitution_default(self):
        """
        Test the functionality of the charmatrix parameter where it does not
        change a character score.
        """
        charmatrix = {('a', 'b'): 1}
        expected = numpy.array([[0, 1],
                                [1, 1]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('a',
                                                                           'b',
                                                                           charmatrix=charmatrix)[0]))

    def test_charmatrix_default_delete(self):
        """
        Test the case where the charmatrix mutates the score of trivial (first
        column) deletes.
        """
        charmatrix = {('', 'a'): 5}
        expected = numpy.array([[0],
                                [5],
                                [10],
                                [15]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('aaa',
                                                                           '',
                                                                           charmatrix=charmatrix)[0]))

    def test_charmatrix_default_insert(self):
        """
        Test the case where the charmatrix mutates the score of trivial (first
        row) inserts.
        """
        charmatrix = {('a', ''): 5}
        expected = numpy.array([[0, 5, 10, 15]])
        self.assertTrue(numpy.array_equal(expected,
                                          self.string.np_full_edit_distance('',
                                                                           'aaa',
                                                                           charmatrix=charmatrix)[0]))
    # -------------------------------------------------------------------
    # Word tests --------------------------------------------------------
    # -------------------------------------------------------------------

    def test_identical_word(self):
        self.assertEqual(0, self.string.np_full_edit_distance(['word'],
                                                             ['word'])[0][-1,
                                                                          -1])

    def test_word_substitution(self):
        self.assertEqual(1, self.string.np_full_edit_distance(['word'],
                                                             ['different\
                                                              word'])[0][-1,
                                                                         -1])

    def test_word_insertion(self):
        self.assertEqual(1, self.string.np_full_edit_distance([],
                                                             ['word'])[0][-1,
                                                                          -1])

    def test_word_deletion(self):
        self.assertEqual(1, self.string.np_full_edit_distance(['word'],
                                                             [])[0][-1, -1])

    def test_word_multi_insert(self):
        self.assertEqual(5, self.string.np_full_edit_distance([], ['word',
                                                                  'word',
                                                                  'word',
                                                                  'word',
                                                                  'word'])[0][-1,
                                                                              -1])

    def test_word_multi_delete(self):
        self.assertEqual(5, self.string.np_full_edit_distance(
            ['word', 'word', 'word', 'word', 'word'], [])[0][-1, -1])

    def test_word_multi_match(self):
        self.assertEqual(0, self.string.np_full_edit_distance(
            ['word1', 'word2', 'word3'], ['word1', 'word2', 'word3'])[0][-1, -1])

    def test_word_multi_substitute(self):
        self.assertEqual(3, self.string.np_full_edit_distance(
            ['word1', 'word2', 'word3'], ['otherword1', 'otherword2', 'otherword3'])[0][-1, -1])


class NumpyAlignmentTests(unittest.TestCase):

    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions


    def test_empty_strings(self):
        """
        Tests the edit steps between two empty strings.
        """
        self.assertEqual([], self.string.np_align('', ''))

    def test_insertions(self):
        """
        Test the edit steps when only inserts are needed.
        """
        self.assertEqual(['i', 'i', 'i', 'i', 'i'],
                         self.string.np_align('', 'abcde'))

    def test_deletions(self):
        """
        Test the edit steps when only deletes are needed.
        """
        self.assertEqual(['d', 'd', 'd', 'd', 'd'],
                         self.string.np_align('abcde', ''))

    def test_matches(self):
        """
        Test the edit steps when only matches are needed.
        """
        self.assertEqual(['m', 'm', 'm', 'm', 'm'],
                         self.string.np_align('abcde', 'abcde'))

    def test_matches_1(self):  # TODO
        """
        Test the edit steps when only substitutions are needed.
        """
        self.assertEqual(['s', 's', 's', 's', 's'],
                         self.string.np_align('abcde', 'vwxyz'))

    def test_wikipedia_example_1(self):
        """
        Test against the first general example provided at
        http://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm
        """
        self.assertEqual(['s', 'm', 'm', 'm', 's', 'm', 'd'],
                         self.string.np_align('sitting', 'kitten'))

    def test_wikipedia_example_2(self):
        """
        Test against the second general example provided at
        http://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm
        """
        self.assertEqual(['m', 'i', 'i', 'm', 's', 'm', 'm', 'm'],
                         self.string.np_align('sunday', 'saturday'))
    # -------------------------------------------------------------------
    # Word Alignment Tests ----------------------------------------------
    # -------------------------------------------------------------------

    def test_empty_word_list(self):
        """
        Tests the edit steps between two empty lists.
        """
        self.assertEqual([], self.string.np_align([], []))

    def test_word_insertions(self):
        """
        Test the edit steps when only inserts are needed.
        """
        self.assertEqual(['i', 'i', 'i', 'i', 'i'], self.string.np_align(
            [], ['word1', 'word2', 'word3', 'word4', 'word5']))

    def test_word_deletions(self):
        """
        Test the edit steps when only inserts are needed.
        """
        self.assertEqual(['d', 'd', 'd', 'd', 'd'], self.string.np_align(
            ['word1', 'word2', 'word3', 'word4', 'word5'], []))

    def test_word_matches(self):
        """
        Test the edit steps when only matches are needed.
        """
        self.assertEqual(['m', 'm', 'm', 'm', 'm'], self.string.np_align(
            ['word1', 'word2', 'word3', 'word4', 'word5'], ['word1', 'word2', 'word3', 'word4', 'word5']))


class NumpySemiGlobalAlignmentTests(unittest.TestCase):

    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions

    """
    Tests the np_semi_global_align function.
    """

    def test_sequence_length_inversion(self):
        """
        Test that an exception is thrown if the first sequence is > the second.
        """
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.np_semi_global_align, 'abcdefghi', '')
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.np_semi_global_align, 'a', '')
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.np_semi_global_align, ['word'], [])
        self.assertRaises(self.exceptions.NidabaAlgorithmException,
                          self.string.np_semi_global_align, ['word1', 'word2',
                                                            'word3'], ['word'])

    def test_identical(self):
        """
        Test with strings and lists of equal length.
        """
        expected = ['m']
        self.assertEqual(expected, self.string.np_semi_global_align('a', 'a'))
        self.assertEqual(expected, self.string.np_semi_global_align(['a'],
                                                                   ['a']))

    def test_identical_long(self):
        """
        Test with a longer series of matches.
        """
        expected = ['m', 'm', 'm', 'm', 'm']
        self.assertEqual(expected, self.string.np_semi_global_align('abcde',
                                                                   'abcde'))

    def test_identical_words(self):
        """
        Test with a series of of matching words.
        """
        expected = ['m', 'm', 'm', 'm', 'm']
        self.assertEqual(expected, self.string.np_semi_global_align(['word1',
                                                                    'word2',
                                                                    'word3',
                                                                    'word4',
                                                                    'word5'],
                                                                   ['word1',
                                                                    'word2',
                                                                    'word3',
                                                                    'word4',
                                                                    'word5']))

    def test_simple_prefix(self):
        """
        Test with a single match and a skippable prefix.
        """
        expected = ['i', 'i', 'i', 'i', 'm']
        self.assertEqual(expected, self.string.np_semi_global_align('b',
                                                                   'aaaab'))

    def test_prefix_with_trailer(self):
        """
        Test with a single match and a skippable prefix and trailer.
        """
        expected = ['i', 'i', 'i', 'i', 'm']
        self.assertEqual(expected, self.string.np_semi_global_align('b',
                                                                   'aaaabcccc'))

    def test_word_prefix(self):
        """
        Test a list of words with a prefix.
        """
        expected = ['i', 'i', 'm']
        self.assertEqual(expected, self.string.np_semi_global_align(['match'],
                                                                   ['w1', 'w2',
                                                                    'match']))

# ----------------------------------------------------------------------
# Language Tests -------------------------------------------------------
# ----------------------------------------------------------------------


class LanguageTests(unittest.TestCase):

    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions


    def test_inblock(self):
        """
        Tests the inblock function using the printable part of the ascii
        block. This block was chosen arbitrarily.
        """
        asciistr = (u"""!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUV"""
                    """WXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~""")
        for c in asciistr:
            self.assertTrue(self.string.inblock(c, ('!', '~')))

        self.assertFalse(self.string.inblock(unichr(ord('!') - 1), ('!', '~')))
        self.assertFalse(self.string.inblock(unichr(ord('~') + 1), ('!', '~')))

    def test_identify_all_ascii_simple(self):
        """
        Test the text identify method against the ascii unicode block.
        """
        asciistr = (u'ascii', unichr(0), unichr(127))
        ex1 = self.string.identify(u'this is a string of ascii', [asciistr])
        self.assertEqual(ex1, {u'ascii': 25})

    def test_identify_all_ascii(self):
        """
        Test the text identify method against the ascii unicode block.
        """
        asciistr = (u'ascii', unichr(0), unichr(127))

        for i in xrange(0, 128):
            self.assertEqual(
                {u'ascii': 1}, self.string.identify(unichr(i), [asciistr]))

    def test_identify_greek_and_latin(self):
        """
        Test the identify method on a string that contains both greek
        and latin characters.
        """
        greek = (u'Greek Coptic', u'\u0370', u'\u03FF')

        self.assertEqual({greek[0]: len(u'Σωκράτης')},
                         self.string.identify(u'Σωκράτης', [greek]))
        self.assertEqual({greek[0]: len(u'Πλάτων')},
                         self.string.identify(u'Πλάτων', [greek]))

    def test_is_lang_threshhold_check(self):
        """
        Test that an exception is raised if an invalid threshold value
        is entered.
        """
        self.assertRaises(Exception, self.string.islang, None, None, -0.00001)
        self.assertRaises(Exception, self.string.islang, None, None, 1.00001)
        self.assertRaises(Exception, self.string.islang, None, None, 0.0)

    def test_is_lang_100percent_ascii(self):
        """
        Test with a pure ascii self.string.
        """
        asciistr = (u"""!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUV
                     WXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~""")
        self.assertTrue(self.string.islang(asciistr, [self.string.ascii_range]))

    def test_is_lang_100percent_greek(self):
        """
        Test with two strings containing characters from the Greek and
        Coptic block.
        """
        gk1 = u'Σωκράτης'
        gk2 = u'Πλάτων'
        self.assertTrue(
            self.string.islang(gk1, [self.string.greek_coptic_range]))
        self.assertTrue(
            self.string.islang(gk2, [self.string.greek_coptic_range]))

    def test_is_lang_50percent_greekascii(self):
        """
        Test with a string that is half ascii and half Greek and Coptic.
        """
        halfandhalf = u'Πλάτων_ascii'
        self.assertTrue(self.string.islang(halfandhalf,
                                          [self.string.ascii_range],
                                          threshold=0.5))
        self.assertTrue(self.string.islang(halfandhalf,
                                          [self.string.greek_coptic_range],
                                          threshold=0.5))

        self.assertFalse(self.string.islang(halfandhalf,
                                           [self.string.greek_coptic_range],
                                           threshold=0.5000001))
        self.assertFalse(self.string.islang(halfandhalf,
                                           [self.string.greek_coptic_range],
                                           threshold=0.5000001))

    def test_isgreek(self):
        """
        Test the isgreek convenience function.
        """
        gk1 = u'Σωκράτης'
        gk2 = u'Πλάτων'
        self.assertTrue(self.string.isgreek(gk1))
        self.assertTrue(self.string.isgreek(gk2))

    def test_sanitize_strip(self):
        """
        Test that sanitize correctly strips leading and trailing whitespace.
        """
        s = u'\n\t abcde \n\t fghij \n\t'
        self.assertEqual(u'abcde \n\t fghij', self.string.sanitize(s))

    def test_sanitize_NFD_decompose(self):
        """
        Test that sanitize correctly converts to NFD codepoints.
        """
        # Small alpha with a tone mark
        alpha = u'\u03AC'
        self.assertNotEqual(alpha, self.string.sanitize(alpha))
        self.assertEqual(u'\u03B1' + u'\u0301', self.string.sanitize(alpha))

        # Greek small letter upsilon with dialytika and tonos
        upsilon = u'\u03B0'
        self.assertNotEqual(upsilon, self.string.sanitize(upsilon))
        self.assertEqual(u'\u03C5' + u'\u0308' + u'\u0301',
                         self.string.sanitize(upsilon))

    def test_sanitize_NFC(self):
        """
        Test that sanitize correctly converts to NFC codepoints.
        """
        decomp_alpha = u'\u03B1' + u'\u0301'
        self.assertNotEqual(decomp_alpha, self.string.sanitize(decomp_alpha,
                                                              normalization=u'NFC'))
        self.assertEqual(u'\u03AC', self.string.sanitize(decomp_alpha,
                                                        normalization=u'NFC'))

        decomp_upsilon = u'\u03C5' + u'\u0308' + u'\u0301'
        self.assertNotEqual(decomp_upsilon, self.string.sanitize(decomp_upsilon,
                                                                normalization=u'NFC'))
        self.assertEqual(
            u'\u03B0', self.string.sanitize(decomp_upsilon, normalization=u'NFC'))

    def test_sanitize_decode(self):
        """
        Test that strings are converted to unicode correctly.
        """
        utf8 = u'I contain a non-ascii character: \u03B1'.encode(u'utf-8')
        utf16 = u'I contain a non-ascii character: \u03B1'.encode(u'utf-16')
        utf32 = u'I contain a non-ascii character: \u03B1'.encode(u'utf-32')
        expected = u'I contain a non-ascii character: \u03B1'
        self.assertEqual(expected, self.string.sanitize(utf8))
        self.assertEqual(
            expected, self.string.sanitize(utf16, encoding=u'utf-16'))
        self.assertEqual(
            expected, self.string.sanitize(utf32, encoding=u'utf-32'))

    def test_diacritic_strip_combining(self):
        """
        Tests the strip_diacritic fuction on the combining diacritics
        unicode block.
        """
        cd = self.string.combining_diacritical_mark_range
        diacritics = self.string.uniblock(cd[1], cd[2])
        self.assertNotEqual(len(diacritics), 0)
        self.assertEqual(u'', self.string.strip_diacritics(diacritics))

    def test_diacritic_strip_noncombining(self):
        """
        Test the strip_diacritic function on all the diacritics not in
        the combining diacritical range.
        """
        gc_diacritics = u'ͺ΄΅'
        gx_diacritics = u'᾽ι᾿῀῁῍῎῏῝῟῞῭΅`´῾'
        diacritics = gc_diacritics + gx_diacritics
        self.assertNotEqual(len(diacritics), 0)
        self.assertEqual(u'', self.string.strip_diacritics(diacritics))

# ----------------------------------------------------------------------
# Symmetric spell check tests ------------------------------------------
# ----------------------------------------------------------------------


class SymSpellTests(unittest.TestCase):

    def setUp(self):
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions

    def test_strings_by_deletion_1(self):
        """
        Test the strings_by_deletion function with one delete.
        """
        expected = [u'abcd', u'abce', u'abde', u'acde', u'bcde']
        self.assertEqual(expected, self.string.strings_by_deletion(u'abcde', 1))

    def test_strings_by_deletion_2(self):
        """
        Test the strings_by_deletion function with two deletes.
        """
        expected = [u'a', u'e', u'p']
        self.assertEqual(expected, self.string.strings_by_deletion(u'ape', 2))

    def test_strings_bzy_deletion_exceed(self):
        """
        Test the strings_by_deletion function where the number of
        deletes exceeds characters in the self.string.
        """
        self.assertEqual([], self.string.strings_by_deletion(u'aaa', 10))

    def test_sym_suggest_already_word(self):
        """
        Test sym_suggest in the case where the specified string is
        already a word in the dictionary.
        """
        delete_dic = {u'ord': [u'word'], u'wod': [u'word'], u'wor': [u'word'],
                      u'wrd': [u'word'], u'ree': [u'tree'], u'tee': [u'tree'],
                      u'tre': [u'tree']}
        dic = {u'tree', u'word'}
        # delete_dic = {u'word':[u'ord', u'wod', u'wor', u'wrd'],
        #        u'tree':[u'ree', u'tee', u'tre']}

        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'word', dic, delete_dic, 1))
        self.assertEqual(
            [u'tree'], self.string.sym_suggest(u'tree', dic, delete_dic, 1))

    def test_sym_suggest_single_delete(self):
        """
        Test sym_suggest where the string differs from words by a
        single delete.
        """
        delete_dic = {u'ord': [u'word'], u'wod': [u'word'], u'wor': [u'word'],
                      u'wrd': [u'word'], u'ree': [u'tree'], u'tee': [u'tree'],
                      u'tre': [u'tree']}
        dic = {u'tree', u'word'}

        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'ord', dic, delete_dic, 1))
        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'wod', dic, delete_dic, 1))
        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'wor', dic, delete_dic, 1))
        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'wrd', dic, delete_dic, 1))

    def test_sym_suggest_single_insert(self):
        """
        Test sym_suggest where the string differs from words by a single
        insert.
        """
        delete_dic = {u'Xword': [u'word'], u'wXord': [u'word'], u'woXrd':
                      [u'word'], u'worXd': [u'word'], u'wordX': [u'word'],
                      u'Xtree': [u'tree']}
        dic = {u'tree', u'word'}

        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'Xword', dic, delete_dic, 1))
        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'wXord', dic, delete_dic, 1))
        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'woXrd', dic, delete_dic, 1))
        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'worXd', dic, delete_dic, 1))
        self.assertEqual(
            [u'word'], self.string.sym_suggest(u'wordX', dic, delete_dic, 1))
        self.assertEqual(
            [u'tree'], self.string.sym_suggest(u'Xtree', dic, delete_dic, 1))

    def test_sym_suggest_single_substitution(self):
        """
        Test sym_suggest where the string differs from words by a single
        insert.
        """
        delete_dic = {u'Word': [u'word'], u'wOrd': [
            u'word'], u'woRd': [u'word'], u'worD': [u'word']}
        dic = {u'word'}

        self.assertEqual([u'word'], self.string.sym_suggest(u'Word', dic,
                                                           delete_dic, 1))
        self.assertEqual([u'word'], self.string.sym_suggest(u'wOrd', dic,
                                                           delete_dic, 1))
        self.assertEqual([u'word'], self.string.sym_suggest(u'woRd', dic,
                                                           delete_dic, 1))
        self.assertEqual([u'word'], self.string.sym_suggest(u'worD', dic,
                                                           delete_dic, 1))

    def test_string_compare(self):
        """
        Test the string comparison function.
        """
        self.assertEqual(1, self.string.compare_strings(u'a', u'b'))
        self.assertEqual(-1, self.string.compare_strings(u'b', u'a'))
        self.assertEqual(0, self.string.compare_strings(u'a', u'a'))
        self.assertEqual(1, self.string.compare_strings(u'a', u'aaaaa'))
        self.assertEqual(-1, self.string.compare_strings(u'aaaaa', u'a'))

    def test_previous_newline_no_newline(self):
        """
        Test the previous newline function when it should move to the
        beginning of the file.
        """
        df = tempfile.NamedTemporaryFile()
        df.write(u'abc')
        df.seek(0, 0)
        with open(os.path.abspath(df.name), 'r+b') as f:
            mm = mmap.mmap(f.fileno(), 0)
            self.assertEqual(0, self.string.prev_newline(mm, 50))
        df.close()

    def test_previous_newline_single_overflow(self):
        """
        Test the previous newline function.
        """
        df = tempfile.NamedTemporaryFile()
        df.write(u'abc\ndef')
        df.seek(0, 0)
        with open(os.path.abspath(df.name), 'r+b') as f:
            mm = mmap.mmap(f.fileno(), 0)
            # Test the points to the left of the newline character
            mm.seek(0)
            self.assertEqual(0, self.string.prev_newline(mm, 50))
            mm.seek(1)
            self.assertEqual(0, self.string.prev_newline(mm, 50))
            mm.seek(2)
            self.assertEqual(0, self.string.prev_newline(mm, 50))
            mm.seek(3)
            self.assertEqual(0, self.string.prev_newline(mm, 50))

            # Test the points to the right of the newline character
            mm.seek(4)
            self.assertEqual(4, self.string.prev_newline(mm, 50))
            mm.seek(5)
            self.assertEqual(4, self.string.prev_newline(mm, 50))
            mm.seek(6)
            self.assertEqual(4, self.string.prev_newline(mm, 50))
            mm.seek(7)
            self.assertEqual(4, self.string.prev_newline(mm, 50))
        df.close()

    def test_previous_newline_multiple_left(self):
        """
        Test the prev_newline function when it would pass several
        newlines.
        """
        df = tempfile.NamedTemporaryFile()
        df.write(u'abcde\nfghijk\nlmnopq\nrstuv\nwxyz')
        df.seek(0, 0)
        with open(os.path.abspath(df.name), 'r+b') as f:
            mm = mmap.mmap(f.fileno(), 0)
            mm.seek(30)
            self.assertEqual(26, self.string.prev_newline(mm, 50))
        df.close()

    def test_key_for_delete_dict_entry(self):
        """
        Tests the key_for_del_dict_entry function.
        """
        single_entry = u'key\tval1'
        multiple_entries = u'key\tval1 val2 val3'
        self.assertEqual((u'key', u'val1'),
                         self.string.key_for_del_dict_entry(single_entry))
        self.assertEqual((u'key', u'val1 val2 val3'),
                         self.string.key_for_del_dict_entry(multiple_entries))

    def test_key_for_single_word(self):
        """
        Tests the test_key_for_single_word function.
        """
        w1 = u'word1'
        w2 = u'word2'
        w3 = u'word3'
        self.assertEqual(
            (u'word1', u'word1'), self.string.key_for_single_word(w1))
        self.assertEqual(
            (u'word2', u'word2'), self.string.key_for_single_word(w2))
        self.assertEqual(
            (u'word3', u'word3'), self.string.key_for_single_word(w3))

    def test_mmap_bin_search_single(self):
        """
        Tests the mmap_bin_search algorithm with a dictionary with
        one item.
        """
        df = tempfile.NamedTemporaryFile()
        df.write(u'only_entry\tsome_value')
        df.seek(0, 0)
        with open(os.path.abspath(df.name), 'r+b') as f:
            mmap.mmap(f.fileno(), 0)
            expected = u'some_value'
            dpath = os.path.abspath(df.name).decode(u'utf-8')
            self.assertEqual(expected,
                             self.string.mmap_bin_search(u'only_entry', dpath))

    def test_mmap_bin_search_double(self):
        """
        Tests the mmap_bin_search algorithm with a dictionary with
        two items.
        """
        df = tempfile.NamedTemporaryFile()
        df.write(u'first_key\tfirst_value\n')
        df.write(u'second_key\tsecond_value')
        df.seek(0, 0)
        with open(os.path.abspath(df.name), 'r+b') as f:
            mmap.mmap(f.fileno(), 0)
            expected_first = u'first_value'
            expected_second = u'second_value'
            dpath = os.path.abspath(df.name).decode(u'utf-8')
            self.assertEqual(expected_first,
                             self.string.mmap_bin_search(u'first_key', dpath))
            self.assertEqual(expected_second,
                             self.string.mmap_bin_search(u'second_key', dpath))

    def test_mmap_bin_search_general(self):
        """
        Test the mmap_bin_search function in a general case.
        """
        df = tempfile.NamedTemporaryFile()
        df.write('akey\taval\n')
        df.write('bkey\tbval\n')
        df.write('ckey\tcval\n')
        df.write('dkey\tdval\n')
        df.write('ekey\teval\n')
        df.write('fkey\tfval\n')
        df.seek(0, 0)
        dpath = os.path.abspath(df.name).decode(u'utf-8')
        ex_a = u'aval'
        ex_b = u'bval'
        ex_c = u'cval'
        ex_d = u'dval'
        ex_e = u'eval'
        ex_f = u'fval'
        self.assertEqual(ex_a, self.string.mmap_bin_search(u'akey', dpath))
        self.assertEqual(ex_b, self.string.mmap_bin_search(u'bkey', dpath))
        self.assertEqual(ex_c, self.string.mmap_bin_search(u'ckey', dpath))
        self.assertEqual(ex_d, self.string.mmap_bin_search(u'dkey', dpath))
        self.assertEqual(ex_e, self.string.mmap_bin_search(u'ekey', dpath))
        self.assertEqual(ex_f, self.string.mmap_bin_search(u'fkey', dpath))
        self.assertEqual(None, self.string.mmap_bin_search(u'gkey', dpath))

    def test_mmap_simple_single_word(self):
        """
        Test the mmap_bin_search function for accessing words from a
        simple word-per-line dictionary.
        """
        df = tempfile.NamedTemporaryFile()
        df.write('aval\n')
        df.write('bval\n')
        df.write('cval\n')
        df.write('dval\n')
        df.seek(0, 0)
        dpath = os.path.abspath(df.name).decode(u'utf-8')
        ex_a = u'aval'
        ex_b = u'bval'
        ex_c = u'cval'
        ex_d = u'dval'
        self.assertEqual(ex_a,
                         self.string.mmap_bin_search(u'aval', dpath,
                                                    entryparser_fn=self.string.key_for_single_word))
        self.assertEqual(ex_b,
                         self.string.mmap_bin_search(u'bval', dpath,
                                                    entryparser_fn=self.string.key_for_single_word))
        self.assertEqual(ex_c,
                         self.string.mmap_bin_search(u'cval', dpath,
                                                    entryparser_fn=self.string.key_for_single_word))
        self.assertEqual(ex_d,
                         self.string.mmap_bin_search(u'dval', dpath,
                                                    entryparser_fn=self.string.key_for_single_word))


class SpellCheckTests(unittest.TestCase):

    def setUp(self):
        self.temp = tempfile.NamedTemporaryFile()
        self.temp.write(u'123\t1234\n')
        self.temp.write(u'124\t1234\n')
        self.temp.write(u'134\t1234\n')
        self.temp.write(u'234\t1234\n')
        self.temp.write(u'aaaa\taaaaa\n')
        self.temp.write(u'bbbb\tbbbbb\n')
        self.temp.seek(0, 0)

        self.dic = set()
        self.dic.add(u'aaaaa')
        self.dic.add(u'bbbbb')
        self.dic.add(u'1234')
        self.config_mock = MagicMock()
        self.config_mock.nidaba.config.everything.log.return_value = True

        modules = {
            'nidaba.config': self.config_mock.config,
        }

        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()
        from nidaba import nidabaexceptions
        from nidaba.algorithms import string
        self.string = string
        self.exceptions = nidabaexceptions

    def tearDown(self):
        self.temp.close()

    def test_suggest_empty_string_1(self):
        """
        Test the spelling suggestor with an empty input string
        """
        result = self.string.mapped_sym_suggest(u'',
                                               self.temp.name.decode('utf-8'),
                                               self.dic, 1)
        self.assertEqual(len(result[u'ins']), 0)
        self.assertEqual(len(result[u'dels']), 0)
        self.assertEqual(len(result[u'subs']), 0)
        self.assertEqual(len(result[u'ins+dels']), 0)

    def test_suggest_1_insert(self):
        """
        Test the spellchecker in the case of a single insert.
        """
        result_a = self.string.mapped_sym_suggest(u'aaaa',
                                                 self.temp.name.decode(
                                                     'utf-8'),
                                                 self.dic, 1)
        result_b = self.string.mapped_sym_suggest(u'bbbb',
                                                 self.temp.name.decode(
                                                     'utf-8'),
                                                 self.dic, 1)
        self.assertEqual(len(result_a[u'ins']), 1)
        self.assertEqual(len(result_a[u'dels']), 0)
        self.assertEqual(len(result_a[u'subs']), 0)
        self.assertEqual(len(result_a[u'ins+dels']), 0)
        self.assertEqual(len(result_b[u'ins']), 1)
        self.assertEqual(len(result_b[u'dels']), 0)
        self.assertEqual(len(result_b[u'subs']), 0)
        self.assertEqual(len(result_b[u'ins+dels']), 0)

    def test_suggest_1_delete(self):
        """
        Test the spellchecker in the case of a single delete.
        """
        result_a = self.string.mapped_sym_suggest(u'aaXaaa',
                                                 self.temp.name.decode(
                                                     'utf-8'),
                                                 self.dic, 1)
        result_b = self.string.mapped_sym_suggest(u'Xbbbbb',
                                                 self.temp.name.decode(
                                                     'utf-8'),
                                                 self.dic, 1)

        self.assertEqual(len(result_a[u'ins']), 0)
        self.assertEqual(len(result_a[u'dels']), 1)
        self.assertEqual(len(result_a[u'subs']), 0)
        self.assertEqual(len(result_a[u'ins+dels']), 0)

        self.assertEqual(len(result_b[u'ins']), 0)
        self.assertEqual(len(result_b[u'dels']), 1)
        self.assertEqual(len(result_b[u'subs']), 0)
        self.assertEqual(len(result_b[u'ins+dels']), 0)

    def test_suggest_1_substitute(self):
        """
        Test the spellchecker in the case of a single substitution.
        """
        result_a = self.string.mapped_sym_suggest(u'aaXaa',
                                                 self.temp.name.decode(
                                                     'utf-8'),
                                                 self.dic, 1)
        result_b = self.string.mapped_sym_suggest(u'Xbbbb',
                                                 self.temp.name.decode(
                                                     'utf-8'),
                                                 self.dic, 1)

        self.assertEqual(len(result_a[u'ins']), 0)
        self.assertEqual(len(result_a[u'dels']), 0)
        self.assertEqual(len(result_a[u'subs']), 1)
        self.assertEqual(len(result_a[u'ins+dels']), 0)

        self.assertEqual(len(result_b[u'ins']), 0)
        self.assertEqual(len(result_b[u'dels']), 0)
        self.assertEqual(len(result_b[u'subs']), 1)
        self.assertEqual(len(result_b[u'ins+dels']), 0)

    def test_suggestions(self):
        """
        Test the suggestions function.
        """
        orig = u'aaaa'
        s0 = u'aaaa'  # match
        s1 = u'baaa'  # edit distance 1
        s1a = u'caaa'  # edit distance 1; after s1 by alphabetical order
        s2 = u'aabb'  # edit distance 2
        s2a = u'aacc'  # edit distance 2; after s2 by alphabetical order
        s3 = u'cccc'  # edit distance 4
        s4 = u'ccccc'  # edit distance 5
        # s4 inserted twice to increase frequency
        expected = [s0, s1, s1a, s2, s2a, s3, s4, s4]
        self.assertEqual(expected, self.string.suggestions(
            orig, [s4, s4, s2, s2a, s0, s1, s1a, s3]))

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import unittest
import os
import shutil
import tempfile
import os

from lxml import etree
from distutils import spawn
from nose.plugins.skip import SkipTest
from mock import patch, MagicMock

thisfile = os.path.abspath(os.path.dirname(__file__))
resources = os.path.abspath(os.path.join(thisfile, 'resources/ocropus'))


class OcropusTests(unittest.TestCase):

    """
    Tests for python ocropus bindings.
    """

    def setUp(self):
        if None in [spawn.find_executable('ocropus-rpred'),
                    spawn.find_executable('ocropus-gpageseg'),
                    spawn.find_executable('ocropus-hocr')]:
            raise SkipTest
        self.config_mock = MagicMock()
        storage_path = unicode(tempfile.mkdtemp())
        self.config_mock.nidaba_cfg = {
            'storage_path': storage_path,
            'ocropus_models': {'ocropus': ('test', 'en-default.pyrnn.gz')},
            'plugins_load': {}
        }

        self.patches = {
            'nidaba.config': self.config_mock,
        }
        self.patcher = patch.dict('sys.modules', self.patches)
        self.patcher2 = patch('nidaba.storage.nidaba_cfg', self.config_mock.nidaba_cfg)
        self.addCleanup(self.patcher2.stop)
        self.addCleanup(self.patcher.stop)
        self.patcher.start()
        self.patcher2.start()
	self.storage_path = storage_path
        shutil.copytree(resources, self.storage_path + '/test')
        from nidaba.plugins import ocropus
        ocropus.setup()
        self.ocropus = ocropus


    def tearDown(self):
        shutil.rmtree(self.storage_path)

    def test_file_path_correct(self):
        """
        Test that output is placed in the correct directory.
        """
        ocr = self.ocropus.ocr_ocropus.run((('test', 'segmentation.xml'),
                                            ('test', 'image_png.png')),
                                           model='ocropus')
        try:
            parser = etree.HTMLParser()
            etree.parse(open(os.path.join(self.storage_path, *ocr)), parser)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_file_outpath_png(self):
        """
        Test that ocropus creates hocr output for pngs.
        """
        ocr = self.ocropus.ocr_ocropus.run((('test', 'segmentation.xml'),
                                            ('test', 'image_png.png')),
                                           model='ocropus')
        try:
            parser = etree.HTMLParser()
            etree.parse(open(os.path.join(self.storage_path, *ocr)), parser)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_file_outpath_tiff(self):
        """
        Test that ocropus creates hocr output for tiffs.
        """
        ocr = self.ocropus.ocr_ocropus.run((('test', 'segmentation.xml'),
                                            ('test', 'image_tiff.tiff')),
                                           model='ocropus')
        try:
            parser = etree.HTMLParser()
            etree.parse(open(os.path.join(self.storage_path, *ocr)), parser)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')

    def test_file_outpath_jpg(self):
        """
        Test that ocropus creates hocr output for jpgs.
        """
        ocr = self.ocropus.ocr_ocropus.run((('test', 'segmentation.xml'),
                                            ('test', 'image_jpg.jpg')),
                                           model='ocropus')
        try:
            parser = etree.HTMLParser()
            etree.parse(open(os.path.join(self.storage_path, *ocr)), parser)
        except etree.XMLSyntaxError:
            self.fail(msg='The output was not valid html/xml!')


if __name__ == '__main__':
    unittest.main()



"""
nidaba.uzn
~~~~~~~~~~

A simple writer/reader interface for UNLV-style zone files.
"""

from __future__ import absolute_import

import csv

from nidaba.nidabaexceptions import NidabaInputException

class UZNReader(object):
    """
    A reader parsing a UNLV zone file from a file object 'f'.
    """

    def __init__(self, f, **kwds):
        self.reader = csv.reader(f, delimiter=' ', skipinitialspace=True, **kwds)

    def next(self):
        row = self.reader.next()
        if len(row) != 5:
            raise NidabaInputException('Incorrect number of columns')
        coords = [int(s) for s in row[:-1]]
        coordinates = [coords[0], 
                       coords[1], 
                       coords[0] + coords[2], 
                       coords[1] + coords[3]]
        return coordinates + [row[-1]]

    def __iter__(self):
        return self


class UZNWriter(object):
    """
    A class writing a UNLV zone file.
    """

    def __init__(self, f, **kwds):
        self.writer = csv.writer(f, delimiter=' ', **kwds)

    def writerow(self, x0, y0, x1, y1, descriptor='Text'):
        self.writer.writerow([x0, y0, x1 - x0, y1 - y0, descriptor])

    def writerows(self, rows):
        for row in rows:
            self.writerow(*row)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import click

from nidaba import lex

@click.command()
@click.option('--input', help='Input text', type=click.Path(exists=True,
              dir_okay=False), required=True)
@click.option('--del_dict', help='Path to the output deletion dictionary',
              type=click.Path(writable=True, dir_okay=False), required=True)
@click.option('--dictionary', help='Path to the output word list',
              type=click.Path(writable=True, dir_okay=False), required=True)
@click.option('--depth', default=1, help='Maximum precalculated edit distance '
              'in deletion dictionary')
@click.version_option()
def main(input, del_dict, dictionary, depth):
    click.echo('Reading input file\t[', nl=False)
    words = lex.cleanuniquewords(input)
    click.secho(u'\u2713', fg='green', nl=False)
    click.echo(']')
    click.echo('Writing dictionary\t[', nl=False)
    lex.make_dict(dictionary, words)
    click.secho(u'\u2713', fg='green', nl=False)
    click.echo(']')
    click.echo('Writing deletions\t[', nl=False)
    lex.make_deldict(del_dict, words, depth)
    click.secho(u'\u2713', fg='green', nl=False)
    click.echo(']')



# -*- coding: utf-8 -*-
"""
nidaba.tasks.helper
~~~~~~~~~~~~~~~~~~~

A helper class that all nidaba tasks should inherit from to ensure accurate
logging of errors.
"""

from __future__ import absolute_import

from celery import Task
from inspect import getargspec
from redis import WatchError
from nidaba.celery import app
from nidaba.config import Redis

import json

def _redis_set_atomically(batch_id, subtask, key, val):
    """
    Atomically sets a field in the Redis batch object to a value.
    """
    with Redis.pipeline() as pipe:
        while 1:
            try:
                pipe.watch(batch_id)
                batch_struct = json.loads(pipe.get(batch_id))
                pipe.multi()
                batch_struct[subtask][key] = val
                pipe.set(batch_id, json.dumps(batch_struct))
                pipe.execute()
                break
            except WatchError:
                continue


class NidabaTask(Task):
    """
    An abstract class propagating unused function arguments through the
    execution chain. This means that no task should accept arbitrary
    (\*\*kwargs) arguments as they won't be forwarded to the actual function
    and will be retained through the whole chain.
    """
    abstract = True
    acks_late = True

    def __call__(self, *args, **kwargs):

        # if args is a dictionary we merge it into kwargs
        if len(args) == 1 and isinstance(args[0], dict):

            kwargs.update(args[0])
            args = ()
            # after a step the output of all tasks is merged into a single
            # argument (indicated by the doc argument being a list of dicts)
            # which we have to unravel first before it can be used be
            # subsequent tasks.
            if isinstance(kwargs['doc'][0], dict):
                docs = []
                root = set()
                for o in kwargs['doc']:
                    docs.append(tuple(o['doc']))
                kwargs['doc'] = docs

        # and then filter all tracking objects (root document, job id, ...) out
        # again
        fspec = getargspec(self.run)
        nkwargs = {}
        tracking_kwargs = {}
        while kwargs:
            k, v = kwargs.popitem()
            if k in fspec.args:
                nkwargs[k] = v
            else:
                tracking_kwargs[k] = v
        task_id = tracking_kwargs['task_id']
        batch_id = tracking_kwargs['batch_id']
        try:
            _redis_set_atomically(batch_id, task_id, 'state', 'RUNNING')
            ret = super(NidabaTask, self).__call__(*args, **nkwargs)
        except Exception as e:
            _redis_set_atomically(batch_id, task_id, 'errors', (nkwargs, e.message))
            _redis_set_atomically(batch_id, task_id, 'state', 'FAILURE')
            raise
        _redis_set_atomically(batch_id, task_id, 'state', 'SUCCESS')

        if isinstance(ret, dict):
            doc = ret.pop('doc')
            if ret:
                _redis_set_atomically(batch_id, task_id, 'misc', ret)
            ret = doc
        _redis_set_atomically(batch_id, task_id, 'result', ret)
        return ret



# -*- coding: utf-8 -*-
"""
nidaba.tasks.postprocessing
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Various postprocessing tasks that operate upon recognized texts.

"""

from __future__ import absolute_import, unicode_literals

from nidaba import storage
from nidaba import merge_hocr
from nidaba import lex
from nidaba.celery import app
from nidaba.tei import TEIFacsimile
from nidaba.config import nidaba_cfg
from nidaba.tasks.helper import NidabaTask


@app.task(base=NidabaTask, name=u'nidaba.postprocessing.spell_check')
def spell_check(doc, method=u'spell_check', language=u'',
                filter_punctuation=False):
    """
    Adds spelling suggestions to an TEI XML document. 

    Alternative spellings for each segment will be included in a choice
    tagcontaining a series of corr tags with the original segment appearing
    beneath a sic element.  Correct words, i.e. words appearing verbatim in the
    dictionary, are left untouched.

    Args:
        doc (unicode, unicode): The input document tuple.
        method (unicode): The suffix string appended to the output file.
        language (unicode): Identifier defined in the nidaba configuration as a
                            valid dictionary.
        filter_punctuation (bool): Switch to filter punctuation inside
                                   ``seg``
    Returns:
        (unicode, unicode): Storage tuple of the output document
    """
    input_path = storage.get_abs_path(*doc)
    output_path = storage.insert_suffix(input_path, method, language,
                                        unicode(filter_punctuation))
    dictionary = storage.get_abs_path(*nidaba_cfg['lang_dicts'][language]['dictionary'])
    del_dictionary = storage.get_abs_path(*nidaba_cfg['lang_dicts'][language]['deletion_dictionary'])
    with storage.StorageFile(*doc) as fp:
        tei = TEIFacsimile()
        tei.read(fp)
        ret = lex.tei_spellcheck(tei, dictionary, del_dictionary,
                                 filter_punctuation)
    with storage.StorageFile(*storage.get_storage_path(output_path), mode='wb') as fp:
        ret.write(fp)
    return storage.get_storage_path(output_path)


#@app.task(base=NidabaTask, name=u'nidaba.postprocessing.blend_hocr')
#def blend_hocr(doc, method=u'blend_hocr', language=u''):
#    """
#    Blends multiple hOCR files using the algorithm from Bruce Robertsons
#    rigaudon. It requires a working spell checking for the input document's
#    language; otherwise all matched bboxes will be bunched together without any
#    scoring.
#
#    Args:
#        doc [(id, path), ...]: A list of storage module tupels that will be
#        merged into a single output document.
#        language (unicode): Language used for spell-checking based scoring. If
#                            not defined no scoring will be used.
#        method (unicode): The suffix string appended to the output file.
#
#    Returns:
#        (unicode, unicode): Storage tuple of the output document
#    """
#    # create the output document path from the first input document
#    input_path = storage.get_abs_path(*doc[0])
#    output_path = storage.insert_suffix(input_path, method)
#    return merge_hocr.merge(doc, language,
#                            storage.get_storage_path(output_path))



from __future__ import absolute_import

from celery import Celery
from nidaba.config import celery_cfg

app = Celery('nidaba',
             include=['nidaba.tasks'])
app.config_from_object(celery_cfg)

if __name__ == '__main__':
    app.start()



# -*- coding: utf-8 -*-
"""
nidaba.nidabaexceptions
~~~~~~~~~~~~~~~~~~~~~~~

All custom exceptions raised by various nidaba modules and packages. Packages
should always define their exceptions here.
"""

class NidabaConfigException(Exception):

    def __init__(self, message=None):
        Exception.__init__(self, message)


class NidabaPluginException(Exception):

    def __init__(self, message=None):
        Exception.__init__(self, message)


class NidabaLeptonicaException(Exception):

    def __init__(self, message=None):
        Exception.__init__(self, message)


class NidabaAlgorithmException(Exception):

    def __init__(self, message=None):
        Exception.__init__(self, message)


class NidabaTaskException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaTickException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaStepException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaInputException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaNoSuchAlgorithmException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaInvalidParameterException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaTEIException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaTesseractException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaOcropusException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaStorageViolationException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaNoSuchStorageBin(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)


class NidabaTEIException(Exception):

    def __init__(self, status_code):
        self.status_code = status_code
        Exception.__init__(self, status_code)



# -*- coding: utf-8 -*-
"""
nidaba.plugins.ocropus
~~~~~~~~~~~~~~~~~~~~~~

Plugin implementing an interface to the `ocropus
<http://github.com/tmbdev/ocropy>`_ OCR engine.

It requires working ocropus-* tools in your execution path. Please have a look
at the website for installation instructions.

.. important::
    If you are not requiring specific functionality of ocropus please consider
    using the :mod:`kraken <nidaba.plugins.kraken>` plugin. Kraken does not
    require working around oddities in input argument acceptance and is
    generally more well-behaved than ocropus.
"""

from __future__ import absolute_import

import os
import numpy as np

from PIL import Image

from nidaba import storage
from nidaba.tei import TEIFacsimile
from nidaba.config import nidaba_cfg
from nidaba.celery import app
from nidaba.tasks.helper import NidabaTask
from nidaba.nidabaexceptions import NidabaOcropusException
from nidaba.nidabaexceptions import NidabaPluginException


def setup(*args, **kwargs):
    try:
        global ocrolib
        import ocrolib
    except:
        raise NidabaPluginException('Prerequisites for ocropus module not '
                                    'installed.')


class micro_hocr(object):
    """
    A simple class encapsulating hOCR attributes
    """
    def __init__(self):
        self.output = u''

    def __str__(self):
        return self.output

    def add(self, *args):
        if self.output:
            self.output += u'; '
        for arg in args:
            if isinstance(arg, basestring):
                self.output += arg + ' '
            elif isinstance(arg, tuple):
                self.output += u','.join([unicode(v) for v in arg]) + u' '
            else:
                self.output += unicode(arg) + u' '
        self.output = self.output.strip()


@app.task(base=NidabaTask, name=u'nidaba.ocr.ocropus')
def ocr_ocropus(doc, method=u'ocr_ocropus', model=None):
    """
    Runs ocropus on an input document.

    Args:
        doc (unicode, unicode): The input document tuple
        method (unicode): The suffix string appended to all output files
        model (unicode): Identifier for the font model to use

    Returns:
        (unicode, unicode): Storage tuple for the output file

    """
    image_path = storage.get_abs_path(*doc[1])
    segmentation_path = storage.get_abs_path(*doc[0])
    output_path = os.path.splitext(storage.insert_suffix(image_path, method,
                                   model))[0] + '.xml'
    model = storage.get_abs_path(*(nidaba_cfg['ocropus_models'][model]))
    return storage.get_storage_path(ocr(image_path, segmentation_path,
                                        output_path, model))


def ocr(image_path, segmentation_path, output_path, model_path):
    """
    Scan a single image with ocropus.

    Reads a single image file from ```imagepath``` and writes the recognized
    text as a TEI document into output_path.

    Args:
        image_path (unicode): Path of the input file
        segmentation_path (unicode): Path of the segmentation XML file.
        output_path (unicode): Path of the output file
        model_path (unicode): Path of the recognition model. Must be a pyrnn.gz
                             pickle dump interoperable with ocropus-rpred.

    Returns:
        (unicode): A string of the output file that is actually written. As
                   Ocropus rewrites output file paths without notice it may be
                   different from the ```outputfilepath``` argument.

    Raises:
        NidabaOcropusException: Ocropus somehow failed. The error output is
                                contained in the message but as it is de facto
                                unusable as a library it's impossible to deduct
                                the nature of the problem.
    """

    try:
        network = ocrolib.load_object(model_path, verbose=0)
        lnorm = getattr(network, "lnorm")
    except Exception as e:
        raise NidabaOcropusException('Something somewhere broke: ' + e.msg)
    im = Image.open(image_path)

    tei = TEIFacsimile()
    with open(segmentation_path, 'r') as seg_fp:
        tei.read(seg_fp)

    # ocropus is a line recognizer
    tei.clear_graphemes()
    tei.clear_segments()
    # add and scope new responsibility statement
    tei.add_respstmt('ocropus', 'character recognition')
    for box in tei.lines:
        ib = tuple(int(x) for x in box[:-2])
        line = ocrolib.pil2array(im.crop(ib))
        temp = np.amax(line)-line
        temp = temp*1.0/np.amax(temp)
        lnorm.measure(temp)
        line = lnorm.normalize(line, cval=np.amax(line))
        if line.ndim == 3:
            np.mean(line, 2)
        line = ocrolib.lstm.prepare_line(line, 16)
        pred = network.predictString(line)
        pred = ocrolib.normalize_text(pred)
        tei.scope_line(box[4])
        tei.add_graphemes(pred)
    with open(output_path, 'wb') as fp:
        tei.write(fp)
    return output_path



# -*- coding: utf-8 -*-
"""
nidaba.lock
~~~~~~~~~~~

This module contains an NFS-safe locking method that is hopefully interoperable
with anything anybody is going to encounter out there.
"""

from __future__ import absolute_import

import os
import time
import random


class lock(object):
    """
    A global lock implementation for files on the common storage medium. It is
    intended to be mainly used on NFS although it should work on any network
    file system.
    """
    def __init__(self, locked_file):
        """
        Initialize a lock object.

        Args:
            locked_file (unicode): Path to the file to be locked
        """
        self._locked_file = locked_file + u'.lock'
        self._lock_file = unicode(os.uname()[1]) + u'.' + \
            unicode(os.getpid()) + u'.lock'

    def acquire(self):
        """
        Acquires a lock on the selected file. Waits until the lock can be
        acquired.
        """
        while True:
            try:
                os.symlink(self._lock_file, self._locked_file)
                break
            except:
                if os.path.islink(self._locked_file) and \
                   os.readlink(self._locked_file) == self._lock_file:
                    break
            time.sleep(random.random() + 0.001)

    def release(self):
        """
        Releases the lock on the selected file.

        Returns:
            bool: True if the lock has been releases, False otherwise
        """
        if os.path.islink(self._locked_file) and \
           os.readlink(self._locked_file) == self._lock_file:
            try:
                os.remove(self._locked_file)
                return True
            except:
                pass
        return False
